from collections import defaultdict
from datetime import datetime, timedelta
import json
import os
from pathlib import Path
import sched
import threading
import time
from typing import List

from peewee import Model, CharField, SqliteDatabase, DateTimeField

from fusion.libs.entity.change import Change
from fusion.util import get_new_id, current_time, timestamp
from fusion.logging import get_logger
from fusion.libs.channel import Channel

from pamet.model.page import Page
from pamet.storage.base_repository import PametRepository
from pamet.storage.file_system.repository import FSStorageRepository

log = get_logger(__name__)

IGNORE_LOCK = False

RECENT = 'recent'
BACKUP = 'backup'
BACKUPS = 'backups'
CHANGESET = 'changeset'
PERMANENT = 'permanent'

HOUR = 60 * 60
DAY = 24 * HOUR
WEEK = 7 * DAY
LUNAR_MONTH = 4 * WEEK

INTERVAL_FIRST_DAY = 1 * HOUR
INTERVAL_FIRST_MONTH = DAY
INTERVAL_UNTIL_MAX = WEEK
PERMANENT_BACKUP_AGE = 366 * DAY

PROCESS_INTERVAL = 30
PRUNE_INTERVAL = DAY
BACKUP_INTERVAL = 1 * HOUR

TIME_FORMAT = '%Y-%m-%dT%H-%M-%S+%z'


def datetime_from_backup_path(path) -> datetime:
    """Exctracts the timestamp from the path/name of a backup file."""
    parts = path.stem.split('_')
    return datetime.strptime(parts[1], TIME_FORMAT)


def datetimes_from_changeset_path(path: Path) -> datetime:
    """Extracts the datetime (start/end) timestamps from a 'changest' file
    path.
    """
    parts = path.stem.split('_')
    return datetime.fromisoformat(parts[1]), datetime.fromisoformat(parts[2])


def file_timestamp(time: datetime):
    return time.strftime(TIME_FORMAT)


class Backup:
    """A minimal convenince class for backup files."""

    def __init__(self, path: Path):
        self.path: Path = path
        self.datetime = datetime_from_backup_path(path)


class ChangePW(Model):
    page_id = CharField()
    time = DateTimeField()
    json = CharField()

    def to_change(self):
        change_dict = json.loads(self.json)
        change = Change.from_safe_delta_dict(change_dict)
        return change


class AnotherServiceAlreadyRunningException(Exception):
    pass


class FSStorageBackupService:
    """Handles creating backups periodically. It has a
    staggered versioning scheme and the option to keep all changes in between
    backups.

    The service connects to the per-TLA changes channel (i.e. receives
    change-sets on every user action. These changes are stored in the
    thread-safe _change_buffer and periodically saved to disk at a
    page-specific tmp_changeset_path.
     This action (done in self.process_changes), backup_changed_pages and
     prune_all are scheduled and executed in a worker thread.

     Backing up saves the changed pages in the respective backup folders and
     moves the change objects (stored in tmp_changeset_file) to a changeset
     file which has the to/from timestamps in it's name.

     Pruning removes backup files according to a staggered versioning scheme.
    I.e. we keep 1 backup for every hour of the first day, 1 per day for the
    first week and one per week after that. When pruning backups - the
    changeset files inbetween get merged.

    Backups older than the PERMANENT_BACKUP_AGE (1 year) get moved to a yearly
    folder (along with the changeset files).

    The backup service has a locking mechanism with a file to avoid running
    multiple services. But that may change in the future.

    Timestamps of the last pruning and last backup operations are kept as files
    in order to schedule the next respective operation correctly.

    The backups are stored in per-page folders. Inside them old/permanent
    backups and changes are stored in per year folders.
     """

    def __init__(self,
                 backup_folder: Path,
                 repository: PametRepository = None,
                 changeset_channel: Channel = None,
                 record_all_changes: bool = False,
                 process_interval: float = PROCESS_INTERVAL,
                 backup_interval: float = BACKUP_INTERVAL,
                 prune_interval: float = PRUNE_INTERVAL,
                 permanent_backup_age: float = PERMANENT_BACKUP_AGE) -> None:
        self.id = get_new_id()

        self.backup_folder: Path = Path(backup_folder)
        self.repo = repository
        self.changeset_channel = changeset_channel
        self.record_all_changes = record_all_changes
        self.changeset_db = None

        self.process_interval = process_interval
        self.backup_interval = backup_interval
        self.prune_interval = prune_interval
        self.permanent_backup_age = permanent_backup_age

        self._changed_page_ids: set = set()
        self.stop_event = threading.Event()

        self.input_channel_sub = None

        self.worker_thread = threading.Thread()
        self.scheduler = sched.scheduler(time.time, time.sleep)

        self._change_buffer: list[Change] = []
        self.buffer_lock = threading.Lock()

        # Prep folders and changeset DB
        self.pages_backup_data_folder().mkdir(exist_ok=True, parents=True)

        if self.record_all_changes:
            self.changeset_db = SqliteDatabase(self.changeset_sqlite_db_path())
            # If the changeset database doesn't exist - create it
            if not self.changeset_sqlite_db_path().exists():

                with self.changeset_db.bind_ctx([ChangePW]):
                    self.changeset_db.create_tables([ChangePW])

    def changeset_sqlite_db_path(self) -> Path:
        return self.backup_folder / 'all_changes.sqlite3'

    def last_prune_timestamp_path(self) -> Path:
        return self.backup_folder / 'last_prune_timestamp.txt'

    def last_backup_timestamp_path(self) -> Path:
        return self.backup_folder / 'last_backup_timestamp.txt'

    def service_lock_path(self) -> Path:
        return self.backup_folder / 'backup_service_lock'

    def pages_backup_data_folder(self):
        return self.backup_folder / BACKUPS

    def page_backup_folder(self, page_id: str) -> Path:
        return self.pages_backup_data_folder() / page_id

    def tmp_changeset_path(self, page_id) -> Path:
        return self.page_backup_folder(page_id) / 'tmp_changeset.jsonl'

    def recent_backup_path(self, page_id, time: datetime) -> Path:
        name = f'{BACKUP}_{file_timestamp(time)}.json'
        return self.page_backup_folder(page_id) / name

    def permanent_backups_folder(self, page_id, time: datetime) -> Path:
        return self.page_backup_folder(page_id) / str(time.year)

    def permanent_backup_path(self, page_id, time: datetime) -> Path:
        name = f'{BACKUP}_{file_timestamp(time)}.json'
        return self.permanent_backups_folder(page_id, time) / name

    # def changeset_path(self, page_id, from_time: datetime,
    #                    to_time: datetime) -> Path:
    #     from_timestamp = timestamp(from_time)
    #     to_timestamp = timestamp(to_time)
    #     file_name = f'{CHANGESET}_{from_timestamp}_{to_timestamp}.jsonl'
    #     path = self.page_backup_folder(page_id) / file_name
    #     return path

    def per_page_backup_folders(self):
        for page_folder in self.pages_backup_data_folder().iterdir():
            if not page_folder.is_dir():
                continue
            yield page_folder

    def backups_for_page(self, page_id) -> list[Path]:
        backups = []
        for root, dirs, files in os.walk(self.page_backup_folder(page_id)):
            for file in files:
                file = Path(file)
                if file.name.startswith(BACKUP):
                    backups.append(file)

        return sorted(backups, key=lambda b: datetime_from_backup_path(b))

    # def existing_backup_path(self, page_id: str, time: datetime):
    #     recent_backup = self.recent_backup_path(page_id, time)
    #     if recent_backup.exists():
    #         return recent_backup

    #     permanent_backup = self.permanent_backup_path(page_id, time)
    #     if permanent_backup.exists():
    #         return permanent_backup

    #     return None

    def recent_backups_for_page(self, page_id) -> List[Path]:
        page_backup_folder = self.page_backup_folder(page_id)
        if not page_backup_folder.exists():
            return []

        backups = []
        for file in page_backup_folder.iterdir():
            if not file.name.startswith(BACKUP):
                continue
            backups.append(file)

        return sorted(backups, key=lambda b: datetime_from_backup_path(b))

    def changeset_paths_for_page(self, page_id) -> List[Path]:
        page_backup_folder = self.page_backup_folder(page_id)
        if not page_backup_folder.exists():
            return []

        changeset_files = []
        for file in page_backup_folder.iterdir():
            if not file.name.startswith(CHANGESET):
                continue
            changeset_files.append(file)

        return sorted(changeset_files,
                      key=lambda b: datetimes_from_changeset_path(b)[0])

    # def last_backup_time(self, page_id: str) -> datetime:
    #     """Returns the timestamp of the most recent backup for a page as a
    #     datetime object."""

    #     last_timestamp = None
    #     for file in self.recent_backups_for_page(page_id):
    #         backup_dt = datetime_from_backup_path(file)
    #         if not last_timestamp:
    #             last_timestamp = backup_dt
    #         elif last_timestamp < backup_dt:
    #             last_timestamp = backup_dt

    #     return last_timestamp

    def last_backup_for_page(self, page_id) -> Path:
        recent_backups = self.recent_backups_for_page(page_id)
        if recent_backups:
            return recent_backups[-1]

        all_backups = self.backups_for_page(page_id)
        if all_backups:
            return all_backups[-1]

        return None

    def last_backup_time(self, page_id: str) -> datetime:
        """Returns the timestamp of the most recent backup for a page as a
        datetime object."""
        last_backup_path = self.last_backup_for_page(page_id)
        if last_backup_path:
            return datetime_from_backup_path(last_backup_path)
        else:
            return None

    def handle_change_set(self, change_set: list[Change]):
        """Should be connected to the per-TLA channel. Adds the received
        changes to the _change_buffer in a thread safe manner."""
        with self.buffer_lock:
            self._change_buffer.extend(change_set)

    def process_changes(self):
        """Sorts the received changes by page id. Stores the ids of pages with
        changes in order to later do backup """
        with self.buffer_lock:
            changes = self._change_buffer
            self._change_buffer = []

        # Sort the changes by page
        changes_by_page = defaultdict(list)
        for change in changes:
            entity = change.last_state()
            page_id = None
            if isinstance(entity, Page):
                page_id = entity.id
            else:
                page_id = entity.page_id

            if not page_id:
                log.error(f'Change f{change} has no page id')
                continue

            changes_by_page[page_id].append(change)

        # Save the ids of the changed pages, so that we know what to back up
        self._changed_page_ids.update(changes_by_page.keys())

        for page_id, changes in changes_by_page.items():
            json_strings = []
            for change in changes:
                json_str = json.dumps(change.as_safe_delta_dict(),
                                      ensure_ascii=False)
                json_strings.append(json_str + '\n')
            json_str_all = ''.join(json_strings)

            tmp_changeset_path = self.tmp_changeset_path(page_id)
            tmp_changeset_path.parent.mkdir(parents=True, exist_ok=True)
            with open(tmp_changeset_path, 'a') as tmp_changeset_file:
                tmp_changeset_file.write(json_str_all)

    def backup_changed_pages(self):
        """Saves the pages marked as changes (in self._changed_page_ids) as
        backup copies and if configured to do so - also saves the change
        objects in files between backups in order to have a fill history."""

        if not self.repo:
            raise Exception('Cannot backup without a configured repository.')

        self.process_changes()  # In case of last second changes

        # Move the recent changes to backup files
        for page_id in self._changed_page_ids:
            # Serialize the page
            page = self.repo.page(page_id)
            if not page:
                # Assume page has been deleted (we don't track that here)
                continue

            notes = self.repo.notes(page)
            arrows = self.repo.arrows(page)
            page_str = FSStorageRepository.serialize_page(page, notes, arrows)

            # Create the backup file
            now = current_time()
            backup_file_path = self.recent_backup_path(page_id, now)
            backup_file_path.parent.mkdir(parents=True, exist_ok=True)
            if backup_file_path.exists():
                error_msg = f'A file already exists at {backup_file_path}'
                log.warning(error_msg)
            backup_file_path.write_text(page_str)

            # If configured to do so - save the changes
            if not self.record_all_changes:
                continue

            # Read them from the tmp changes file, and add them to the db
            tmp_changeset_file = self.tmp_changeset_path(page_id)
            lines = tmp_changeset_file.read_text().splitlines()
            with self.changeset_db.bind_ctx([ChangePW]):
                with self.changeset_db.atomic():
                    for change_json in lines:
                        change = Change.from_safe_delta_dict(
                            json.loads(change_json))
                        change = ChangePW.create(page_id=page_id,
                                                 time=change.time,
                                                 json=change_json)
            tmp_changeset_file.unlink()

        self._changed_page_ids.clear()
        self.last_backup_timestamp_path().write_text(timestamp(current_time()))

    def move_to_permanent_where_appropriate(self, page_id: str):
        # For those older than self.permanent_backup_age -
        # move them to a yearly folder
        page_folder = self.page_backup_folder(page_id)

        for_permanent_backup = []
        for file_path in page_folder.iterdir():
            if not file_path.name.startswith(BACKUP) or \
                    not file_path.is_file():
                continue

            backup = Backup(file_path)
            backup_age = current_time() - backup.datetime

            if backup_age.total_seconds() > self.permanent_backup_age:
                for_permanent_backup.append(backup)

        # Prep the changes file paths and to_times
        changeset_paths_by_to_times = {}
        for file_path in self.page_backup_folder(page_id).iterdir():
            if file_path.name.startswith(CHANGESET):
                from_time, to_time = datetimes_from_changeset_path(file_path)
                changeset_paths_by_to_times[to_time] = file_path

        # We're assuming they're already cleaned a week apart
        # If not - big deal
        for backup in for_permanent_backup:
            folder = self.permanent_backups_folder(page_id, backup.datetime)
            folder.mkdir(parents=True, exist_ok=True)
            new_path = folder / backup.path.name
            backup.path.rename(new_path)
            log.info(f'Moved backup {backup.path.name} '
                     f'to the yearly folder {folder}')

            # # Also move the corresponding changes file if any
            # changeset_path = changeset_paths_by_to_times.get(backup.datetime)
            # if changeset_path:
            #     changeset_path.rename(folder / changeset_path.name)

    def prune_all(self):
        """Calls the pruning method for all pages."""
        # for page_folder in self.per_page_backup_folders():
        #     self.prune_page_backups(page_folder.name)
        #     self.move_to_permanent_where_appropriate(page_folder.name)
        for page_folder in self.per_page_backup_folders():
            page_id = page_folder.name
            self.prune_page_backups(page_id)

            self.move_to_permanent_where_appropriate(page_id)
        self.last_prune_timestamp_path().write_text(timestamp(current_time()))

    def prune_page_backups(self, page_id):
        """Removes old backups in accordance to the staggered scheme and
        configured intervals for each respective period. By default those are:
        Once per hour for the first day.
        Once per day for the first week.
        Once per week until the end of time.
        Also backups older than the PERMANENT_BACKUP_AGE are considered
        permanent and are moved to a per-year folder. Those are not iterated
        over while pruning.
        """
        now = current_time()

        # Sort the backups into bins according to how old they are
        last_day = []
        last_month = []
        older_than_a_month = []
        page_folder = self.page_backup_folder(page_id)
        for file_path in page_folder.iterdir():
            if not file_path.name.startswith(BACKUP) or \
                    not file_path.is_file():
                continue

            backup = Backup(file_path)
            backup_age = now - backup.datetime
            backup_age = backup_age.total_seconds()

            if backup_age < DAY:
                last_day.append(backup)
            elif backup_age < LUNAR_MONTH:
                last_month.append(backup)
            elif backup_age < self.permanent_backup_age:
                older_than_a_month.append(backup)

        # Bin by hour, day, and week (by the clock/calendar),sort, and
        # delete all except the last for each group
        by_hour = defaultdict(list)
        for backup in last_day:
            by_hour[backup.datetime.hour].append(backup)
        by_day = defaultdict(list)
        for backup in last_month:
            by_day[backup.datetime.day].append(backup)
        by_year_and_week = defaultdict(list)
        for backup in older_than_a_month:
            year, week, _ = backup.datetime.isocalendar()
            by_year_and_week[(year, week)].append(backup)

        for_removal = []
        for_keeping = []
        for by_interval in [by_hour, by_day, by_year_and_week]:
            for specific_time_frame, backups in by_interval.items():
                if not backups:
                    continue

                if len(backups) > 1:
                    # Sort chronologically
                    backups = sorted(backups, key=lambda b: b.datetime)

                    # Remove all except the last
                    for_removal.extend(backups[:-1])

                for_keeping.append(backups[-1])

        # Delete the backups files marked for removal
        for backup in for_removal:
            backup.path.unlink()
            log.info(f'Pruned backup {backup.path}')

    def backup_and_reschedule(self):
        self.backup_changed_pages()
        self.scheduler.enter(self.backup_interval, 2,
                             self.backup_and_reschedule)

    def prune_and_reschedule(self):
        self.prune_all()
        self.scheduler.enter(self.prune_interval, 2, self.prune_and_reschedule)

    def process_changes_and_reschedule(self):
        self.process_changes()
        self.scheduler.enter(self.process_interval, 1,
                             self.process_changes_and_reschedule)

    def run_scheduler(self):
        """Execute scheduled operations.

        This gets called in a separate thread. Continues until the stop_event
        is set."""
        timeout = 1
        while timeout and timeout > 0:
            timeout = self.scheduler.run(blocking=False)
            self.stop_event.wait(timeout)

    def start(self):
        """Starts the backup service.

        The startup process includes checking for a service lock, connecting to
        the changes channel, reading the backup storage for per-page changes
        that have not yet been backed up, and reading the last backup and
        prune timestamps.
        The latter are used to schedule backup and pruning operations at the
        time.

        Then the worker thread is started to carry out event scheduling and
        rescheduling."""
        # if self.service_lock_path().exists() and not IGNORE_LOCK:
        #     raise AnotherServiceAlreadyRunningException  #@IgnoreException
        # self.service_lock_path().write_text(self.id)

        log.info('Starting')

        # Setup the receival of changes from the change channel if one is given
        if self.changeset_channel:
            self.input_channel_sub = self.changeset_channel.subscribe(
                self.handle_change_set)

        # Search for tmp_changeset files to populate the _changed_page_ids.
        # I.e. figure out which pages expect a backup because some changes
        # were made before the scheduled backup
        for page_backup_folder in self.per_page_backup_folders():
            page_id = page_backup_folder.name
            if self.tmp_changeset_path(page_id).exists():
                self._changed_page_ids.add(page_id)

        # Get the timestamp for the last backup and schedule accordingly
        backup_ts = self.last_backup_timestamp_path()
        if backup_ts.exists():
            last_backup_dt = datetime.fromisoformat(backup_ts.read_text())
            now = current_time()

            # Check if it's time to do a backup
            time_since_last_backup = now - last_backup_dt
            if time_since_last_backup.total_seconds() > self.backup_interval:
                self.backup_and_reschedule()
            else:
                backup_delta = timedelta(seconds=self.backup_interval)
                time_till_next_backup = backup_delta - time_since_last_backup
                self.scheduler.enter(time_till_next_backup.total_seconds(), 2,
                                     self.backup_and_reschedule)
        else:  # First run
            self.backup_and_reschedule()

        # Get the timestamp for the last pruning and schedule accordingly
        prune_ts = self.last_prune_timestamp_path()
        if prune_ts.exists():
            last_prune_dt = datetime.fromisoformat(prune_ts.read_text())
            now = current_time()

            # Check if it's time to do prunung
            time_since_last_prune = now - last_prune_dt
            if time_since_last_prune.total_seconds() > self.prune_interval:
                self.prune_and_reschedule()
            else:
                prune_delta = timedelta(seconds=self.prune_interval)
                time_till_next_prune = prune_delta - time_since_last_prune
                self.scheduler.enter(time_till_next_prune.total_seconds(), 2,
                                     self.prune_and_reschedule)
        else:  # First run
            self.prune_and_reschedule()

        # Schedule the processing of changes
        self.scheduler.enter(self.process_interval, 1,
                             self.process_changes_and_reschedule)

        self.worker_thread = threading.Thread(target=self.run_scheduler)
        self.worker_thread.start()

    def stop(self):
        """Stops the worker thread and unsubscribes from changes."""
        if self.input_channel_sub:
            self.input_channel_sub.unsubscribe()

        for event in self.scheduler.queue:
            self.scheduler.cancel(event=event)
        log.info('Cancelled events.')

        self.stop_event.set()
        self.worker_thread.join()

        # if self.service_lock_path().exists():
        #     self.service_lock_path().unlink()
        log.info('Stopped service.')

    def get_changes(self,
                    page_id: str = None,
                    after_time: datetime = None,
                    before_time: datetime = None):
        with self.changeset_db.bind_ctx([ChangePW]):
            where_args = []
            if page_id is not None:
                where_args.append(ChangePW.page_id == page_id)
            if after_time is not None:
                where_args.append(ChangePW.time >= after_time)
            if before_time is not None:
                where_args.append(ChangePW.time < before_time)

            query = ChangePW.select()
            if where_args:
                query = query.where(*where_args)

            for change_pw in query.iterator():
                yield change_pw.to_change()



import pamet
from fusion.visual_inspection.action_camera import exec_action
from fusion import fsm
from pamet.model.arrow import ArrowAnchorType

from pamet.model.text_note import TextNote
from fusion.util.point2d import Point2D

# Create a second note
# Create arrows with free and anchored ends
# Also create one that starts and ends on the same note
# Also try one that starts and ends on the same anchor
# Update control points
# Move notes while arrows present
# Delete arrows


def test_arrow_crud(window_fixture, request):
    run_headless = request.config.getoption('--headless')

    with exec_action(delay_before_next=0.5032045841217041,
                     apply_delay=not run_headless):
        pamet.actions.map_page.clear_child_selection(
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.note.create_new_note(
            tab_state=fsm.view_state('2'),
            note=TextNote(id=("00000001", "00000005"),
                          geometry=[570.0, -371.0, 320, 160],
                          style={},
                          content={"text": ""},
                          metadata={},
                          created="2022-09-09T10:00:35+03:00",
                          modified="2022-09-09T10:00:35+03:00",
                          tags=[]))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.note.finish_creating_note(
            tab_state=fsm.view_state('2'),
            note=TextNote(id=("00000001", "00000005"),
                          geometry=[570.0, -371.0, 320, 160],
                          style={},
                          content={"text": ""},
                          metadata={},
                          created="2022-09-09T10:00:37+03:00",
                          modified="2022-09-09T10:00:37+03:00",
                          tags=[]))

    with exec_action(delay_before_next=0.7165634632110596,
                     apply_delay=not run_headless):
        pamet.actions.map_page.start_arrow_creation(
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.515120267868042,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(-112.0, -567.0),
            anchor_note_id=None,
            anchor_type=ArrowAnchorType.NONE,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.010453462600708008,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(-100.0, -553.0),
            anchor_note_id=None,
            anchor_type=ArrowAnchorType.NONE,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.005595207214355469,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_click(
            fixed_pos=Point2D(-100.0, -553.0),
            anchor_note_id=None,
            anchor_type=ArrowAnchorType.NONE,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(74.0, -601.0),
            anchor_note_id=None,
            anchor_type=ArrowAnchorType.NONE,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.5248458385467529,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(496.0, -673.0),
            anchor_note_id=None,
            anchor_type=ArrowAnchorType.NONE,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_click(
            fixed_pos=Point2D(496.0, -673.0),
            anchor_note_id=None,
            anchor_type=ArrowAnchorType.NONE,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.5091159343719482,
                     apply_delay=not run_headless):
        pamet.actions.map_page.start_arrow_creation(
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.5119011402130127,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(184.0, -65.0),
            anchor_note_id="00000002",
            anchor_type=ArrowAnchorType.MID_RIGHT,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.5125882625579834,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(136.0, -41.0),
            anchor_note_id="00000002",
            anchor_type=ArrowAnchorType.MID_RIGHT,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.5067262649536133,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(-16.0, -73.0),
            anchor_note_id="00000002",
            anchor_type=ArrowAnchorType.TOP_MID,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.006644010543823242,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(-8.0, -83.0),
            anchor_note_id="00000002",
            anchor_type=ArrowAnchorType.TOP_MID,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.005877256393432617,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_click(
            fixed_pos=Point2D(-8.0, -83.0),
            anchor_note_id="00000002",
            anchor_type=ArrowAnchorType.TOP_MID,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(162.0, -185.0),
            anchor_note_id=None,
            anchor_type=ArrowAnchorType.NONE,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.8889937400817871,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(568.0, -309.0),
            anchor_note_id="00000005",
            anchor_type=ArrowAnchorType.MID_LEFT,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.5109269618988037,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(568.0, -307.0),
            anchor_note_id="00000005",
            anchor_type=ArrowAnchorType.MID_LEFT,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.005484580993652344,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(568.0, -301.0),
            anchor_note_id="00000005",
            anchor_type=ArrowAnchorType.MID_LEFT,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_click(
            fixed_pos=Point2D(568.0, -301.0),
            anchor_note_id="00000005",
            anchor_type=ArrowAnchorType.MID_LEFT,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.5188963413238525,
                     apply_delay=not run_headless):
        pamet.actions.map_page.start_arrow_creation(
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.5179836750030518,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(752.0, -379.0),
            anchor_note_id="00000005",
            anchor_type=ArrowAnchorType.TOP_MID,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.527916669845581,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(746.0, -381.0),
            anchor_note_id="00000005",
            anchor_type=ArrowAnchorType.TOP_MID,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.533435583114624,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(738.0, -381.0),
            anchor_note_id="00000005",
            anchor_type=ArrowAnchorType.TOP_MID,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.9121277332305908,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_click(
            fixed_pos=Point2D(738.0, -381.0),
            anchor_note_id="00000005",
            anchor_type=ArrowAnchorType.TOP_MID,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.5174984931945801,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(736.0, -383.0),
            anchor_note_id="00000005",
            anchor_type=ArrowAnchorType.TOP_MID,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.6047213077545166,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(720.0, -389.0),
            anchor_note_id="00000005",
            anchor_type=ArrowAnchorType.TOP_MID,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.5158109664916992,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(720.0, -391.0),
            anchor_note_id="00000005",
            anchor_type=ArrowAnchorType.TOP_MID,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_click(
            fixed_pos=Point2D(720.0, -391.0),
            anchor_note_id="00000005",
            anchor_type=ArrowAnchorType.TOP_MID,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.7181804180145264,
                     apply_delay=not run_headless):
        pamet.actions.map_page.start_arrow_creation(
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.5064895153045654,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(436.0, -431.0),
            anchor_note_id=None,
            anchor_type=ArrowAnchorType.NONE,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.5123739242553711,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(696.0, -233.0),
            anchor_note_id="00000005",
            anchor_type=ArrowAnchorType.BOTTOM_MID,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.5148122310638428,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(720.0, -211.0),
            anchor_note_id="00000005",
            anchor_type=ArrowAnchorType.BOTTOM_MID,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.016089916229248047,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(726.0, -209.0),
            anchor_note_id="00000005",
            anchor_type=ArrowAnchorType.BOTTOM_MID,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_click(
            fixed_pos=Point2D(726.0, -209.0),
            anchor_note_id="00000005",
            anchor_type=ArrowAnchorType.BOTTOM_MID,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.5145213603973389,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(380.0, -47.0),
            anchor_note_id=None,
            anchor_type=ArrowAnchorType.NONE,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.6868588924407959,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(216.0, -3.0),
            anchor_note_id=None,
            anchor_type=ArrowAnchorType.NONE,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.5117301940917969,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(218.0, -3.0),
            anchor_note_id=None,
            anchor_type=ArrowAnchorType.NONE,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.9635162353515625,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(376.0, -29.0),
            anchor_note_id=None,
            anchor_type=ArrowAnchorType.NONE,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_click(
            fixed_pos=Point2D(376.0, -29.0),
            anchor_note_id=None,
            anchor_type=ArrowAnchorType.NONE,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.5073676109313965,
                     apply_delay=not run_headless):
        pamet.actions.map_page.clear_child_selection(
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.note.create_new_note(
            tab_state=fsm.view_state('2'),
            note=TextNote(id=("00000001", "00000017"),
                          geometry=[566.0, 203.0, 320, 160],
                          style={},
                          content={"text": ""},
                          metadata={},
                          created="2022-09-09T10:01:10+03:00",
                          modified="2022-09-09T10:01:10+03:00",
                          tags=[]))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.note.finish_creating_note(
            tab_state=fsm.view_state('2'),
            note=TextNote(id=("00000001", "00000017"),
                          geometry=[566.0, 203.0, 320, 160],
                          style={},
                          content={"text": ""},
                          metadata={},
                          created="2022-09-09T10:01:11+03:00",
                          modified="2022-09-09T10:01:11+03:00",
                          tags=[]))

    with exec_action(delay_before_next=0.002137899398803711,
                     apply_delay=not run_headless):
        pamet.actions.map_page.clear_child_selection(
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.update_child_selections(
            map_page_view_state=fsm.view_state('3'),
            selection_updates_by_child={fsm.view_state('19'): True})

    with exec_action(delay_before_next=0.5176196098327637,
                     apply_delay=not run_headless):
        pamet.actions.map_page.start_arrow_creation(
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.5153346061706543,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(574.0, 293.0),
            anchor_note_id="00000017",
            anchor_type=ArrowAnchorType.MID_LEFT,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.5281515121459961,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(542.0, 291.0),
            anchor_note_id="00000017",
            anchor_type=ArrowAnchorType.MID_LEFT,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.013361454010009766,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(558.0, 285.0),
            anchor_note_id="00000017",
            anchor_type=ArrowAnchorType.MID_LEFT,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_click(
            fixed_pos=Point2D(558.0, 285.0),
            anchor_note_id="00000017",
            anchor_type=ArrowAnchorType.MID_LEFT,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.9202549457550049,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_move(
            fixed_pos=Point2D(0.0, 93.0),
            anchor_note_id="00000002",
            anchor_type=ArrowAnchorType.BOTTOM_MID,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.arrow_creation_click(
            fixed_pos=Point2D(0.0, 93.0),
            anchor_note_id="00000002",
            anchor_type=ArrowAnchorType.BOTTOM_MID,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.012984037399291992,
                     apply_delay=not run_headless):
        pamet.actions.map_page.clear_child_selection(
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.update_child_selections(
            map_page_view_state=fsm.view_state('3'),
            selection_updates_by_child={fsm.view_state('22'): True})

    with exec_action(delay_before_next=0.5147323608398438,
                     apply_delay=not run_headless):
        pamet.actions.map_page.start_arrow_edge_drag(
            map_page_view_state=fsm.view_state('3'),
            mouse_pos=Point2D(1034, 600),
            edge_index=0.5)

    with exec_action(delay_before_next=0.8191120624542236,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_edge_drag_update(
            fixed_pos=Point2D(422.0, 125.0),
            anchor_note_id=None,
            anchor_type=ArrowAnchorType.NONE,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.9684457778930664,
                     apply_delay=not run_headless):
        pamet.actions.map_page.finish_arrow_edge_drag(
            fixed_pos=Point2D(422.0, 125.0),
            anchor_note_id=None,
            anchor_type=ArrowAnchorType.NONE,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.5132126808166504,
                     apply_delay=not run_headless):
        pamet.actions.map_page.start_arrow_edge_drag(
            map_page_view_state=fsm.view_state('3'),
            mouse_pos=Point2D(1005, 546),
            edge_index=1.5)

    with exec_action(delay_before_next=0.8163204193115234,
                     apply_delay=not run_headless):
        pamet.actions.map_page.arrow_edge_drag_update(
            fixed_pos=Point2D(146.0, 275.0),
            anchor_note_id=None,
            anchor_type=ArrowAnchorType.NONE,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.finish_arrow_edge_drag(
            fixed_pos=Point2D(146.0, 275.0),
            anchor_note_id=None,
            anchor_type=ArrowAnchorType.NONE,
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.003901958465576172,
                     apply_delay=not run_headless):
        pamet.actions.map_page.clear_child_selection(
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.update_child_selections(
            map_page_view_state=fsm.view_state('3'),
            selection_updates_by_child={fsm.view_state('19'): True})

    with exec_action(delay_before_next=0.5120954513549805,
                     apply_delay=not run_headless):
        pamet.actions.map_page.start_child_move(
            map_page_view_state=fsm.view_state('3'),
            mouse_pos=Point2D(1299, 621))

    with exec_action(delay_before_next=0.5062975883483887,
                     apply_delay=not run_headless):
        pamet.actions.map_page.moved_child_view_update(
            map_page_view_state=fsm.view_state('3'),
            delta=Point2D(68.0, -50.0))

    with exec_action(delay_before_next=0.9184830188751221,
                     apply_delay=not run_headless):
        pamet.actions.map_page.moved_child_view_update(
            map_page_view_state=fsm.view_state('3'),
            delta=Point2D(132.0, -96.0))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.finish_child_move(
            map_page_view_state=fsm.view_state('3'),
            delta=Point2D(132.0, -96.0))

    with exec_action(delay_before_next=0.003981828689575195,
                     apply_delay=not run_headless):
        pamet.actions.map_page.clear_child_selection(
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.update_child_selections(
            map_page_view_state=fsm.view_state('3'),
            selection_updates_by_child={fsm.view_state('6'): True})

    with exec_action(delay_before_next=0.5107238292694092,
                     apply_delay=not run_headless):
        pamet.actions.map_page.start_child_move(
            map_page_view_state=fsm.view_state('3'),
            mouse_pos=Point2D(1299, 328))

    with exec_action(delay_before_next=0.5067813396453857,
                     apply_delay=not run_headless):
        pamet.actions.map_page.moved_child_view_update(
            map_page_view_state=fsm.view_state('3'),
            delta=Point2D(104.0, -6.0))

    with exec_action(delay_before_next=0.5138816833496094,
                     apply_delay=not run_headless):
        pamet.actions.map_page.moved_child_view_update(
            map_page_view_state=fsm.view_state('3'),
            delta=Point2D(186.0, -8.0))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.finish_child_move(
            map_page_view_state=fsm.view_state('3'),
            delta=Point2D(186.0, -8.0))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.delete_selected_children(
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.0014564990997314453,
                     apply_delay=not run_headless):
        pamet.actions.map_page.clear_child_selection(
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.update_child_selections(
            map_page_view_state=fsm.view_state('3'),
            selection_updates_by_child={fsm.view_state('4'): True})

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.delete_selected_children(
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.0010895729064941406,
                     apply_delay=not run_headless):
        pamet.actions.map_page.clear_child_selection(
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.update_child_selections(
            map_page_view_state=fsm.view_state('3'),
            selection_updates_by_child={fsm.view_state('9'): True})

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.delete_selected_children(
            map_page_view_state=fsm.view_state('3'))

    assert True



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'widget.ui'
##
## Created by: Qt User Interface Compiler version 6.3.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QHBoxLayout, QLineEdit, QPushButton,
    QSizePolicy, QVBoxLayout, QWidget)

class Ui_TabMainWidget(object):
    def setupUi(self, TabMainWidget):
        if not TabMainWidget.objectName():
            TabMainWidget.setObjectName(u"TabMainWidget")
        TabMainWidget.resize(1019, 625)
        self.horizontalLayout = QHBoxLayout(TabMainWidget)
        self.horizontalLayout.setSpacing(6)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.leftSidebarContainer = QWidget(TabMainWidget)
        self.leftSidebarContainer.setObjectName(u"leftSidebarContainer")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(5)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.leftSidebarContainer.sizePolicy().hasHeightForWidth())
        self.leftSidebarContainer.setSizePolicy(sizePolicy)
        self.leftSidebarContainer.setMinimumSize(QSize(100, 0))
        self.verticalLayout = QVBoxLayout(self.leftSidebarContainer)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.leftSidebarCloseButton = QPushButton(self.leftSidebarContainer)
        self.leftSidebarCloseButton.setObjectName(u"leftSidebarCloseButton")

        self.verticalLayout.addWidget(self.leftSidebarCloseButton)


        self.horizontalLayout.addWidget(self.leftSidebarContainer)

        self.centralContainer = QWidget(TabMainWidget)
        self.centralContainer.setObjectName(u"centralContainer")
        sizePolicy1 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        sizePolicy1.setHorizontalStretch(30)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.centralContainer.sizePolicy().hasHeightForWidth())
        self.centralContainer.setSizePolicy(sizePolicy1)
        self.verticalLayout_3 = QVBoxLayout(self.centralContainer)
        self.verticalLayout_3.setSpacing(6)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.commandLineEdit = QLineEdit(self.centralContainer)
        self.commandLineEdit.setObjectName(u"commandLineEdit")

        self.verticalLayout_3.addWidget(self.commandLineEdit)


        self.horizontalLayout.addWidget(self.centralContainer)

        self.rightSidebarContainer = QWidget(TabMainWidget)
        self.rightSidebarContainer.setObjectName(u"rightSidebarContainer")
        sizePolicy2 = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Preferred)
        sizePolicy2.setHorizontalStretch(5)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.rightSidebarContainer.sizePolicy().hasHeightForWidth())
        self.rightSidebarContainer.setSizePolicy(sizePolicy2)
        self.rightSidebarContainer.setMinimumSize(QSize(100, 0))
        self.verticalLayout_2 = QVBoxLayout(self.rightSidebarContainer)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.rightSidebarCloseButton = QPushButton(self.rightSidebarContainer)
        self.rightSidebarCloseButton.setObjectName(u"rightSidebarCloseButton")

        self.verticalLayout_2.addWidget(self.rightSidebarCloseButton)


        self.horizontalLayout.addWidget(self.rightSidebarContainer)


        self.retranslateUi(TabMainWidget)

        QMetaObject.connectSlotsByName(TabMainWidget)
    # setupUi

    def retranslateUi(self, TabMainWidget):
        TabMainWidget.setWindowTitle(QCoreApplication.translate("TabMainWidget", u"Form", None))
        self.leftSidebarCloseButton.setText(QCoreApplication.translate("TabMainWidget", u"<", None))
        self.rightSidebarCloseButton.setText(QCoreApplication.translate("TabMainWidget", u">", None))
    # retranslateUi




from PySide6.QtCore import QUrl
from PySide6.QtGui import QDesktopServices
from fusion.util.point2d import Point2D
from fusion.view import View

import pamet
from pamet.actions import window as window_actions
from pamet.actions import tab as tab_actions
from pamet.actions import note as note_actions


class NoteView(View):
    recieves_double_click_events: bool = False

    def __init__(self, parent, initial_state):
        View.__init__(self, initial_state=initial_state)
        self.parent_page = parent

    def left_mouse_double_click_event(self, position: Point2D):
        url = self.state().url
        if url.is_empty() or url.is_custom():
            note_actions.start_editing_note(
                self.parent_page.parent_tab.state(),
                self.state().get_note())
        else:
            self.left_mouse_double_click_with_link(position)

    def left_mouse_double_click_with_link(self, position: Point2D):
        state = self.state()

        # If there's a linked page - go to it
        if state.url.is_internal():
            page = pamet.page(state.url.get_page_id())
            if page:
                tab_actions.go_to_url(self.parent_page.parent_tab.state(),
                                      page.url())
            else:
                super().left_mouse_double_click_event(position)
        # IMPLEMENT opening no schema urls (non-page names) and http/https
        else:
            QDesktopServices.openUrl(QUrl(str(state.url)))

    def middle_click_event(self, position: Point2D):
        url = self.state().url
        page = pamet.page(url.get_page_id())

        if not page:
            return

        window_view = self.parent_page.parent_tab.parent_window
        window_actions.new_browser_tab(window_view.state(), page)



from copy import copy
from typing import Generator, List, Union

import fusion
from fusion.util import Point2D, Rectangle
from fusion.view import View
import pamet

from pamet.constants import ARROW_EDGE_RAIDUS, RESIZE_CIRCLE_RADIUS, ARROW_SELECTION_RADIUS
from pamet.desktop_app.util import control_is_pressed, shift_is_pressed
from pamet.constants import MOVE_SPEED, MIN_HEIGHT_SCALE, MAX_HEIGHT_SCALE

from pamet import actions
from pamet.actions import map_page as map_page_actions
from pamet.actions import tab as tab_actions
from pamet.model.arrow import ArrowAnchorType
from pamet.model.text_note import TextNote
from pamet.views.arrow.widget import ArrowView
from pamet.views.map_page.state import MapPageViewState, MapPageMode
from pamet.views.note.base.view import NoteView

log = fusion.get_logger(__name__)


class MapPageView(View):

    def __init__(self, parent, initial_state):
        super().__init__(initial_state=initial_state)
        self.parent_tab = parent

        self._state = initial_state
        self._left_mouse_is_pressed = False
        self._mouse_position_on_left_press = Point2D(0, 0)
        self._note_views_under_mouse = []

    def note_views(self) -> List[NoteView]:
        raise NotImplementedError

    def arrow_views(self) -> List[ArrowView]:
        raise NotImplementedError

    @property
    def page(self):
        return self.state().page.copy()

    @property
    def viewport(self):
        return self.state().viewport

    def get_note_view_state_for_note_gid(self, note_gid: str):
        for nv_state in self.state().note_view_states:
            if nv_state.note_gid == note_gid:
                return nv_state
        return None

    def get_note_views_in_area(self, rect: Rectangle):
        unprojected_rect = self.state().unproject_rect(rect)
        intersecting = []

        for child in self.note_views():
            if child.state().rect().intersects(unprojected_rect):
                intersecting.append(child)

        return intersecting

    def get_note_views_at(
            self,
            position: Point2D,
            radius: float = 0) -> Generator[NoteView, None, None]:
        unprojected_mouse_pos = self.state().unproject_point(position)

        for note_view in self.note_views():
            nv_rect = note_view.state().rect()
            if radius:
                # Instead of implementing a Rectangle.intersects(circle)
                # we expand the rectangle to do the same inference
                nv_rect: Rectangle = copy(nv_rect)
                nv_rect.set_size(nv_rect.size() +
                                 Point2D(radius * 2, radius * 2))
                nv_rect.set_top_left(nv_rect.top_left() -
                                     Point2D(radius, radius))

            if nv_rect.contains(unprojected_mouse_pos):
                yield note_view

    def get_note_view_at(self,
                         position: Point2D,
                         radius: float = 0) -> Union[NoteView, None]:
        for note_view in self.get_note_views_at(position, radius):
            return note_view
        return None

    def get_arrow_views_at(
            self, position: Point2D) -> Generator[ArrowView, None, None]:
        unprojected_pos = self.state().unproject_point(position)
        for arrow_view in self.arrow_views():
            if arrow_view.intersects_circle(unprojected_pos,
                                            ARROW_SELECTION_RADIUS):
                yield arrow_view

    def get_arrow_view_at(self, position: Point2D) -> Union[ArrowView, None]:
        for arrow_view in self.get_arrow_views_at(position):
            return arrow_view
        return None

    def resize_circle_intersect(self, position: Point2D):
        for nc in self.note_views():
            unprojected_pos = self.state().unproject_point(position)
            resize_circle_center = nc.state().rect().bottom_right()

            distance = resize_circle_center.distance_to(unprojected_pos)
            if distance <= RESIZE_CIRCLE_RADIUS:
                return nc

    def arrow_anchor_at(self, position: Point2D):
        """Returns a tuple (note_id, anchor_type) for the closest anchor"""
        unprojected_pos = self.state().unproject_point(position)

        # Get the note views close enough to the mouse to show anchors
        adjacent_note_views = self.get_note_views_at(position,
                                                     ARROW_EDGE_RAIDUS)

        # Get the note, anchor, distance tuples of the intersections
        intersections = []
        for note_view in adjacent_note_views:
            for anchor_type in ArrowAnchorType.real_types():
                note_vs = note_view.state()
                anchor_center = note_vs.arrow_anchor(anchor_type)
                distance = unprojected_pos.distance_to(anchor_center)
                intersections.append((note_view, anchor_type, distance))

        if not intersections:
            return None

        # Sort them and return the first
        intersections = sorted(intersections, key=lambda i: i[2])
        note_view, anchor_type = intersections[0][:2]
        return note_view.state().get_note().own_id, anchor_type

    def handle_delete_shortcut(self):
        map_page_actions.delete_selected_children(self.state())

    def handle_left_mouse_long_press(self, mouse_pos: Point2D):
        if self.state().mode() != MapPageMode.NONE:
            return

        nc_under_mouse = self.get_note_view_at(mouse_pos)
        arrow_view_under_mouse = self.get_arrow_view_at(mouse_pos)
        child_under_mouse = nc_under_mouse

        # If there's a note - prefer ignore the arrow
        if arrow_view_under_mouse and not nc_under_mouse:
            # Ignore  arrows that are not free floating on both sides
            av_state = arrow_view_under_mouse.state()
            if av_state.tail_coords and av_state.head_coords:
                child_under_mouse = arrow_view_under_mouse

        if child_under_mouse:
            state = self.state()
            if child_under_mouse.state() not in state.selected_children:
                map_page_actions.update_child_selections(
                    state, {child_under_mouse.state(): True})
            map_page_actions.start_child_move(self.state(), mouse_pos)

    def handle_left_mouse_press(self, mouse_pos: Point2D):
        self._mouse_position_on_left_press = copy(mouse_pos)
        self._left_mouse_is_pressed = True

        ctrl_pressed = control_is_pressed()
        shift_pressed = shift_is_pressed()
        state = self.state()
        unproj_mouse_pos = state.unproject_point(mouse_pos)

        child_under_mouse = self.get_arrow_view_at(mouse_pos)
        nv_under_mouse = self.get_note_view_at(mouse_pos)
        if nv_under_mouse:
            child_under_mouse = nv_under_mouse
        resize_nv = self.resize_circle_intersect(mouse_pos)

        if state.mode() == MapPageMode.CREATE_ARROW:
            anchor_props = self.arrow_anchor_at(mouse_pos)
            note_id, anchor_type = anchor_props or (None, ArrowAnchorType.NONE)
            map_page_actions.arrow_creation_click(state,
                                                  fixed_pos=unproj_mouse_pos,
                                                  anchor_note_id=note_id,
                                                  anchor_type=anchor_type)
            return

        if ctrl_pressed and shift_pressed:
            map_page_actions.start_drag_select(state, mouse_pos)
            return

        if ctrl_pressed and not shift_pressed:
            if child_under_mouse:
                nc_selected = child_under_mouse.state() in self.state(
                ).selected_children
                map_page_actions.update_child_selections(
                    state, {child_under_mouse.state(): not nc_selected})
        elif shift_pressed or (ctrl_pressed and shift_pressed):
            # Select the note under the mouse (if it's not already so)
            if child_under_mouse:
                nc_selected = child_under_mouse.state() in self.state(
                ).selected_children
                if not nc_selected:
                    map_page_actions.update_child_selections(
                        state, {child_under_mouse.state(): True})

        # Clear selection (or reduce it to the note under the mouse)
        if not ctrl_pressed and not shift_pressed:
            if resize_nv:
                child_under_mouse = resize_nv

            map_page_actions.clear_child_selection(state)

            if child_under_mouse:
                map_page_actions.update_child_selections(
                    state, {child_under_mouse.state(): True})

        # Check for resize initiation
        if resize_nv:
            resize_nv_state = resize_nv.state()
            if resize_nv_state not in state.selected_children:
                map_page_actions.update_child_selections(
                    state, {resize_nv_state: True})

            resize_circle_center = resize_nv_state.rect().bottom_right()
            rcc_projected = state.project_point(resize_circle_center)
            map_page_actions.start_notes_resize(state,
                                                resize_nv_state.get_note(),
                                                mouse_pos, rcc_projected)

    def handle_left_mouse_release(self, mouse_pos: Point2D):
        self._left_mouse_is_pressed = False

        state: MapPageViewState = self.state()
        mode = state.mode()
        unproj_mouse_pos = state.unproject_point(mouse_pos)

        if mode == MapPageMode.DRAG_SELECT:
            same_pos = state.mouse_position_on_drag_select_start == mouse_pos
            map_page_actions.stop_drag_select(state, mouse_pos)
            # If no drag select has happened, but it was just a click - add
            # the note under the mouse to the selection.
            if same_pos:
                child_under_mouse = self.get_arrow_view_at(mouse_pos)
                nv_under_mouse = self.get_note_view_at(mouse_pos)
                if nv_under_mouse:
                    child_under_mouse = nv_under_mouse
                map_page_actions.update_child_selections(
                    state, {child_under_mouse.state(): True})

        elif mode == MapPageMode.NOTE_RESIZE:
            new_size = self._new_note_size_on_resize(mouse_pos)
            map_page_actions.finish_notes_resize(self.state(), new_size)

        elif mode == MapPageMode.CHILD_MOVE:
            pos_delta = mouse_pos - state.mouse_position_on_note_drag_start
            pos_delta /= self.state().height_scale_factor()
            map_page_actions.finish_child_move(self.state(), pos_delta)

        elif mode == MapPageMode.DRAG_NAVIGATION:
            map_page_actions.finish_mouse_drag_navigation(self.state())
            tab_actions.update_current_url(self.parent_tab.state())

        elif mode == MapPageMode.ARROW_EDGE_DRAG:
            anchor_props = self.arrow_anchor_at(mouse_pos)
            note_id, anchor_type = anchor_props or (None, ArrowAnchorType.NONE)
            map_page_actions.finish_arrow_edge_drag(state,
                                                    fixed_pos=unproj_mouse_pos,
                                                    anchor_note_id=note_id,
                                                    anchor_type=anchor_type)

    def _new_note_size_on_resize(self, new_mouse_pos: Point2D) -> Point2D:
        mouse_delta = new_mouse_pos - self.state().note_resize_click_position
        size_delta = mouse_delta - self.state(
        ).note_resize_delta_from_note_edge

        size_delta = size_delta / self.state().height_scale_factor()
        new_size = self.state().note_resize_main_note.size() + size_delta

        return new_size

    def _handle_move_on_drag_select(self, mouse_pos: Point2D):
        selection_rect = Rectangle.from_points(
            self.state().mouse_position_on_drag_select_start, mouse_pos)
        unprojected_rect = self.state().unproject_rect(selection_rect)

        nvs_in_selection = self.get_note_views_in_area(selection_rect)
        arrow_views_in_selection = [
            av for av in self.arrow_views()
            if av.intersects_rect(unprojected_rect)
        ]
        drag_selected_children = set([
            child.state()
            for child in nvs_in_selection + arrow_views_in_selection
        ])

        map_page_actions.update_drag_select(self.state(),
                                            selection_rect.as_tuple(),
                                            drag_selected_children)

    def handle_mouse_move(self, mouse_pos: Point2D):
        state = self.state()
        mode = state.mode()
        delta = self._mouse_position_on_left_press - mouse_pos
        unproj_mouse_pos = state.unproject_point(mouse_pos)
        self._note_views_under_mouse = list(self.get_note_views_at(mouse_pos))

        if mode == MapPageMode.NONE:
            if self._left_mouse_is_pressed:
                map_page_actions.start_mouse_drag_navigation(
                    state, self._mouse_position_on_left_press, delta)

        if mode == MapPageMode.DRAG_SELECT:
            self._handle_move_on_drag_select(mouse_pos)

        elif mode == MapPageMode.NOTE_RESIZE:
            new_size = self._new_note_size_on_resize(mouse_pos)
            map_page_actions.resize_note_views(state, new_size)

        elif mode == MapPageMode.CHILD_MOVE:
            pos_delta = mouse_pos - self._mouse_position_on_left_press
            pos_delta /= state.height_scale_factor()
            map_page_actions.moved_child_view_update(state, pos_delta)

        elif mode == MapPageMode.DRAG_NAVIGATION:
            map_page_actions.mouse_drag_navigation_move(state, delta)

        elif mode == MapPageMode.CREATE_ARROW:
            anchor_props = self.arrow_anchor_at(mouse_pos)
            note_id, anchor_type = anchor_props or (None, ArrowAnchorType.NONE)
            map_page_actions.arrow_creation_move(state,
                                                 fixed_pos=unproj_mouse_pos,
                                                 anchor_note_id=note_id,
                                                 anchor_type=anchor_type)

        elif mode == MapPageMode.ARROW_EDGE_DRAG:
            anchor_props = self.arrow_anchor_at(mouse_pos)
            note_id, anchor_type = anchor_props or (None, ArrowAnchorType.NONE)
            map_page_actions.arrow_edge_drag_update(state,
                                                    fixed_pos=unproj_mouse_pos,
                                                    anchor_note_id=note_id,
                                                    anchor_type=anchor_type)

    def handle_mouse_scroll(self, steps: int):
        delta = MOVE_SPEED * steps
        current_height = self.state().viewport_height

        new_height = max(MIN_HEIGHT_SCALE,
                         min(current_height - delta, MAX_HEIGHT_SCALE))

        map_page_actions.set_viewport_height(self.state(), new_height)
        tab_actions.update_current_url(self.parent_tab.state())

    def left_mouse_double_click_event(self, mouse_pos: Point2D):
        if self.state().mode() == MapPageMode.CREATE_ARROW:
            map_page_actions.abort_special_mode(self.state())
            return

        note_view = self.get_note_view_at(mouse_pos)

        if note_view:
            # Map the mouse position on the note and call its virtual method
            note_rect: Rectangle = note_view.state().get_note().rect()
            mouse_pos_unproj = self.state().unproject_point(mouse_pos)
            relative_pos = note_rect.top_left() - mouse_pos_unproj
            note_view.left_mouse_double_click_event(relative_pos)
        else:
            note_pos = self.state().unproject_point(mouse_pos)
            note = TextNote().with_id(page_id=self.state().page_id)
            note.x = note_pos.x()
            note.y = note_pos.y()
            actions.note.create_new_note(self.parent_tab.state(), note)

    def middle_click_event(self, mouse_pos: Point2D):
        note_view = self.get_note_view_at(mouse_pos)

        if note_view:
            note_view.middle_click_event(mouse_pos)

    def handle_resize_event(self, new_size: Point2D):
        map_page_actions.resize_page(self.state(), new_size)

    def undo(self):
        pamet.undo_service().back_one_step(self.state().page_id)

    def redo(self):
        pamet.undo_service().forward_one_step(self.state().page_id)



from ast import Tuple
from collections import defaultdict
from copy import copy
# from hashlib import md5
import json

from datetime import datetime, timedelta
from pathlib import Path
import shutil
from typing import List
from fusion.libs.entity import load_from_dict

from fusion.util import Point2D
from fusion.util.rectangle import Rectangle
from fusion.util import current_time, get_new_id, timestamp
from fusion.logging import get_logger
from fusion.storage.in_memory_repository import InMemoryRepository
from pamet.constants import MAX_NOTE_HEIGHT, MAX_NOTE_WIDTH, MIN_NOTE_HEIGHT
from pamet.constants import MIN_NOTE_WIDTH
from pamet.util import snap_to_grid
from pamet.model.page import Page
from pamet.model.arrow import Arrow, ArrowAnchorType
from pamet.model.image_note import ImageNote
from pamet.model.note import Note
from pamet.model.script_note import ScriptNote
from pamet.model.text_note import TextNote

log = get_logger(__name__)

_cache = {}
_paths_cache = {}
MAX_CACHE = 1000
new_to_old_path = {}
md5_by_old_name = {}

# # Caching - to be removed. it's only for the bin import
# def get_from_cache(page_path: Path):
#     if page_md5 in _cache:
#         return _cache[page_md5]
#     return page_md5

TIME_FORMAT = '%d.%m.%Y %H:%M:%S'
ONE_V3_COORD_UNIT_TO_V4 = 20

INTERNAL_ANCHOR_PREFIX = 'this_note_points_to:'
EXTERNAL_ANCHOR_PREFIX = 'define_web_page_note:'
IMAGE_NOTE_PREFIX = 'define_picture_note:'
SYSTEM_CALL_NOTE_PREFIX = 'define_system_call_note:'


def backup_file(file_path: Path, backup_folder: Path):
    backup_folder.mkdir(parents=True, exist_ok=True)
    backup_path = backup_folder / (file_path.name + '.backup')
    if backup_path.exists():
        backup_name = (backup_path.stem + f'.backup-{get_new_id()}')
        backup_path = backup_folder / backup_name

    shutil.copy(file_path, backup_path)
    log.info(f'Backed up file {file_path} to {backup_path}')

    return backup_path


def new_legacy_id_for_legacy_note(note_id: int, timestamp: str, text: str,
                                  all_ids: list):
    # Make a deterministic new id based on the timestamp and old id
    new_id = hash(timestamp)
    if new_id in all_ids:
        # If that's not enough - add a suffix based on the note text
        new_id = hash(timestamp + text)
        if new_id in all_ids:  # Mostly for timeline notes
            new_id = hash(get_new_id())
            if new_id in all_ids:
                raise Exception('WTF')

    return new_id


def new_id_for_legacy_note(note_id, timestamp, content: str, all_ids: list):
    # Make a deterministic new id based on the timestamp and old id
    note_id = str(note_id)
    timestamp = str(timestamp)
    new_id = get_new_id(note_id + timestamp)
    if new_id in all_ids:
        # If that's not enough - add a suffix based on the note text
        new_id = get_new_id(note_id + timestamp + str(content))
        if new_id in all_ids:  # Mostly for timeline notes
            raise Exception('WTF')

    return new_id


class LegacyFSRepoReader:

    def __init__(self) -> None:
        self.v2_note_checksum_by_page_name = {}
        self.v3_note_checksum_by_page_name = {}
        self.v2_notes_by_page_name = defaultdict(set)

    def convert_v3_to_v4(self, json_path: str | Path, backup_folder: Path,
                         previous_v_repo_entities: dict):
        # V3 example: {
        # "is_displayed_first_on_startup": true
        # "notes": [
        # {"bg_col": [0,0,1,0.1],
        #     "font_size": 1,
        #     "height": 3,
        #     "id": 1,
        #     "links": [
        #         {
        #             "cp": [
        #                 12.25,
        #                 -7.25
        #             ],
        #             "text": "",
        #             "to_id": 1
        #         }
        #     ],
        #     "t_made": "10.4.2019 14:2:2",
        #     "t_mod": "10.4.2019 14:2:2",
        #     "tags": [],
        #     "text": "this_note_points_to:notes",
        #     "txt_col": [0,0,1,1],
        #     "width": 11,
        #     "x": 1,
        #     "y": -2.5}]}

        json_path = Path(json_path)
        page_data = json.loads(json_path.read_text())
        notes_data = page_data.pop('notes')

        # Load the page
        page = Page(name=json_path.stem, id=get_new_id(json_path.stem))
        self.v3_note_checksum_by_page_name[page.name] = 0

        # Load the notes and arrows
        earliest_creation_time = current_time()
        notes = []
        arrows = []
        notes_by_id = {}
        ids_with_duplicates = []
        new_ids_by_old = {}
        for nt in notes_data:
            self.v3_note_checksum_by_page_name[page.name] += 1
            # Scale old coords so that default widget fonts
            # look adequate without correction
            for coord in ['x', 'y', 'width', 'height']:
                nt[coord] = snap_to_grid(nt[coord] * ONE_V3_COORD_UNIT_TO_V4)

                # There was that very specific case with an overflow, wtf
                if not (-2147483647 < nt[coord] < 2147483647):
                    nt[coord] = 1

            # nt['page_id'] = page.id

            nt['id'] = (page.id, str(nt['id']))
            nt['color'] = nt.pop('txt_col')
            nt['background_color'] = nt.pop('bg_col')
            nt['geometry'] = [
                nt.pop('x'),
                nt.pop('y'),
                nt.pop('width'),
                nt.pop('height')
            ]
            t_made = nt.pop('t_made')
            t_mod = nt.pop('t_mod')

            if not t_made:
                log.warning('In page {page.name}: Note {nt} is missing t_made')
                t_made = datetime.strftime(current_time(), TIME_FORMAT)
            if not t_mod:
                log.warning('In page {page.name}: Note {nt} is missing t_mod')
                t_mod = datetime.strftime(current_time(), TIME_FORMAT)

            created = datetime.strptime(t_made, TIME_FORMAT)
            modified = datetime.strptime(t_mod, TIME_FORMAT)
            # Fuckit, use the local timezone
            created = created.astimezone()
            modified = modified.astimezone()
            nt['created'] = timestamp(created)
            nt['modified'] = timestamp(modified)

            nt.pop('font_size', None)

            if created < earliest_creation_time:
                earliest_creation_time = created

            if not page.id:
                raise Exception

            # Arrows
            legacy_arrows = nt.pop('links')
            for legacy_arrow in legacy_arrows:
                arrow = Arrow.in_page(page)
                arrow.set_tail(anchor_note_id=nt['id'][1])
                arrow.set_head(anchor_note_id=str(legacy_arrow['to_id']))
                control_point = legacy_arrow.pop('cp', None)
                if control_point:
                    midpoint = Point2D(*control_point)
                    midpoint *= ONE_V3_COORD_UNIT_TO_V4
                    arrow.replace_midpoints([midpoint])

                arrows.append(arrow)

            # Import according to the note type
            note_text = nt.pop('text')

            # Internal anchors
            if note_text.startswith(INTERNAL_ANCHOR_PREFIX):
                page_name = note_text[len(INTERNAL_ANCHOR_PREFIX):]
                nt['type_name'] = TextNote.__name__
                note = load_from_dict(nt)
                note.url = page_name  # Will be converted to a url afterwards

            # External anchors (web links)
            elif note_text.startswith(EXTERNAL_ANCHOR_PREFIX):
                lines = note_text.split('\n')
                url = ''
                text = ''
                for line in lines:
                    if line.startswith('url='):
                        url = line[4:]
                    elif line.startswith('name='):
                        text = line[5:]

                nt['type_name'] = TextNote.__name__
                note = load_from_dict(nt)
                note.url = url
                note.text = text

            # Image notes
            elif note_text.startswith(IMAGE_NOTE_PREFIX):
                image_url = note_text[len(IMAGE_NOTE_PREFIX):]
                nt['type_name'] = ImageNote.__name__
                note = load_from_dict(nt)
                note.image_url = image_url
                note.local_image_url = image_url

            # System call notes
            elif note_text.startswith(SYSTEM_CALL_NOTE_PREFIX):
                script = note_text[len(SYSTEM_CALL_NOTE_PREFIX):]
                script_parts = script.split()
                command_args = ''.join(script_parts[1:])

                nt['type_name'] = ScriptNote.__name__
                note = load_from_dict(nt)
                note.script_path = script_parts[0]
                note.command_args = command_args
                note.text = script

            else:  # It's just a text note
                nt['type_name'] = TextNote.__name__
                note = load_from_dict(nt)
                note.text = note_text

            # Detect duplicate ids
            if note.own_id in notes_by_id:
                ids_with_duplicates.append(note.own_id)
                log.warning(f'Detected duplicate for {note}. '
                            f'Links to/from it will be deleted.')
                # f'Url: {note.url}, text: {note.text}. |'
                # f'Duplicate {note}'
                # f'Url: {notes_by_id[note.own_id].url}, '
                # f'text: {notes_by_id[note.own_id].text}.')

            old_id = note.own_id
            note = note.with_id(own_id=new_id_for_legacy_note(
                note.own_id,
                note.created,
                note.content,
                notes_by_id.keys(),
            ))
            new_ids_by_old[old_id] = note.own_id

            # Check note sizes
            size = note.rect().size()
            size_x = max(min(size.x(), MAX_NOTE_WIDTH), MIN_NOTE_WIDTH)
            size_y = max(min(size.y(), MAX_NOTE_HEIGHT), MIN_NOTE_HEIGHT)
            if size_x != size.x() or size_y != size.y():
                log.info(f'Note with invalid size imported. '
                         f'Old: {size.x(), size.y()}; New: {size_x, size_y}')
                rect = note.rect()
                rect.set_size(Point2D(size_x, size_y))
                note.set_rect(rect)

            notes.append(note)
            notes_by_id[note.own_id] = note

        # Remove arrows which start or end at notes with duplicate ids, because
        # we don't want to deal with that at all
        arrows = [
            a for a in arrows if arrow.tail_note_id not in ids_with_duplicates
            and arrow.head_note_id not in ids_with_duplicates
        ]

        for arrow in arrows:
            # Fix arrow ids (since we changed the note ids)
            arrow.tail_note_id = new_ids_by_old[arrow.tail_note_id]
            arrow.head_note_id = new_ids_by_old[arrow.head_note_id]

            tail_note = notes_by_id[arrow.tail_note_id]
            head_note = notes_by_id[arrow.head_note_id]

            tail_rect = tail_note.rect()
            head_rect: Rectangle = head_note.rect()

            # To check if the notes are above one another - move to the same
            # height and check for an intersection
            intersect_check_rect = copy(head_rect)
            intersect_check_rect.set_y(tail_rect.y())

            if tail_rect.intersects(intersect_check_rect):
                if head_rect.center().y() < tail_rect.center().y():
                    arrow.tail_anchor_type = ArrowAnchorType.TOP_MID
                    arrow.head_anchor_type = ArrowAnchorType.BOTTOM_MID
                else:
                    arrow.tail_anchor_type = ArrowAnchorType.BOTTOM_MID
                    arrow.head_anchor_type = ArrowAnchorType.TOP_MID
            elif head_rect.right() <= tail_rect.left():
                arrow.tail_anchor_type = ArrowAnchorType.MID_LEFT
                arrow.head_anchor_type = ArrowAnchorType.MID_RIGHT
            elif head_rect.left() >= tail_rect.right():
                arrow.tail_anchor_type = ArrowAnchorType.MID_RIGHT
                arrow.head_anchor_type = ArrowAnchorType.MID_LEFT
            else:
                raise Exception

        # Update the arrow ids to be deterministic
        arrows_with_new_ids = []
        arrows_by_id = {}
        for arrow in arrows:
            new_id = get_new_id([arrow.tail_note_id, arrow.head_note_id])
            if new_id in arrows_by_id:
                # This means there's a second arrow with the same start and
                # end. We just remove it.
                continue
            arrows_by_id[new_id] = arrow
            arrows_with_new_ids.append(arrow.with_id(arrow.page_id, new_id))
        arrows = arrows_with_new_ids

        # Assume page created time from the notes. Modified makes sense only
        # for renames as of now and there's no basis to infer it
        page.datetime_created = earliest_creation_time - timedelta(seconds=10)
        page.datetime_modified = earliest_creation_time - timedelta(seconds=10)

        new_path = self.path_for_page(page)
        page_json_str = self.serialize_page(page, notes, arrows)

        # Disabled on DEBUG for now
        # if new_path.exists():
        #     raise Exception or move or something

        new_path.write_text(page_json_str)

        backup_file(json_path, backup_folder)
        log.info(f'Converted V3 page {json_path} to V4 at {new_path} ')
        json_path.unlink()

        return new_path

    def convert_v2_to_v3(self, file_path: str | Path, backup_folder: Path):
        # Example:
        # [79367]
        # txt=this_note_points_to:Програмиране
        # x=-7
        # y=-4.5
        # z=0
        # a=9
        # b=2.5
        # font_size=1
        # t_made=6.10.2019 14:1:12
        # t_mod=6.10.2019 14:1:12
        # txt_col=0;0;1;1
        # bg_col=0;0;1;0.100008
        # l_id=
        # l_txt=
        # l_CP_x=
        # l_CP_y=
        # tags=

        file_path = Path(file_path)
        self.v2_note_checksum_by_page_name[file_path.stem] = 0

        misl_file_string = file_path.read_text()

        is_displayed_first_on_startup = False
        is_a_timeline_notefile = False

        # Clear the windows standart junk
        # The \r chars should be handled by python
        # misl_file_string = misl_file_string.replace('\r', '')

        lines = [ln for ln in misl_file_string.split('\n') if ln]  # Skip empty

        lines_left = []
        for line in lines:
            if line.startswith('#'):  # Skip comments
                continue

            elif line.startswith('is_displayed_first_on_startup'):
                is_displayed_first_on_startup = True
                continue

            elif line.startswith('is_a_timeline_note_file'):
                is_a_timeline_notefile = True
                continue

            lines_left.append(line)

        if is_a_timeline_notefile:
            return

        # Extract groups
        notes = defaultdict(dict)
        current_note_id = None
        changed_note_ids = []
        meta_for_id_replace = {}  # Gets assigned the target id
        for line in lines_left:

            if line.startswith('[') and line.endswith(']'):
                current_note_id = int(line[1:-1])

                if current_note_id in self.v2_notes_by_page_name[
                        file_path.stem]:
                    changed_note_ids.append(current_note_id)
                    old_id = current_note_id
                    current_note_id = get_new_id()
                    if current_note_id in meta_for_id_replace:
                        Exception('Not enough randomness, wtf')
                    meta_for_id_replace[current_note_id] = old_id

                self.v2_note_checksum_by_page_name[file_path.stem] += 1
                self.v2_notes_by_page_name[file_path.stem].add(current_note_id)
                continue

            if '=' not in line:
                raise Exception

            [key, value] = line.split('=', 1)

            # Drop some unused fields
            if key in ['z', 'l_txt']:
                continue

            # Convert values where needed
            elif key == 'txt':
                value = value.replace('\\n', '\n')
            elif key in ['txt_col', 'bg_col']:
                value = [float(channel) for channel in value.split(';')]
            elif key in ['l_id', 'l_CP_x', 'l_CP_y']:
                value = value.split(';')
                value = [v for v in value if v]
                if key == 'l_id':
                    value = [int(v) for v in value]
                else:
                    value = [float(v) for v in value]
            elif key == 'id':
                value = int(value)
            elif key in ['x', 'y', 'a', 'b']:
                value = float(value)
            elif key in ['tags', 'font_size', 't_mod', 't_made']:
                pass
            else:
                raise Exception

            notes[current_note_id][key] = value

        # Notes with duplicate ids first get a random id and then we
        # replace it with a deterministic id based on
        # id, timestamp, text
        for nt_random_id, old_id in meta_for_id_replace.items():
            nt = notes.pop(nt_random_id)
            note_id = new_legacy_id_for_legacy_note(old_id, nt['t_made'],
                                                    nt['txt'],
                                                    list(notes.keys()))
            notes[note_id] = nt

        for note_id, nt in notes.items():
            # Rename fields to the V3 schema
            nt['id'] = note_id
            nt['width'] = nt.pop('a')
            nt['height'] = nt.pop('b')
            nt['text'] = nt.pop('txt')

            # Process arrows (formerly named links)
            control_points = []
            if 'l_CP_x' in nt and 'l_CP_y' in nt:
                for cp_x, cp_y in zip(nt['l_CP_x'], nt['l_CP_y']):
                    control_points.append(Point2D(cp_x, cp_y))
            else:
                control_points = None

            to_ids = nt.pop('l_id')
            cp_xs = nt.pop('l_CP_x', None)
            cp_ys = nt.pop('l_CP_y', None)

            if cp_xs and cp_ys:
                CPs = list(zip(cp_xs, cp_ys))
                # # Clear empty ones
                # CPs = [cp for cp in CPs if cp[0] and cp[1]]
                if len(to_ids) != len(CPs):
                    raise Exception
            else:
                CPs = None

            nt['links'] = []
            for i, to_id in enumerate(to_ids):
                if to_id in changed_note_ids:  # Skip links of buggy notes
                    continue
                link_dict = {'text': '', 'to_id': to_id}
                if CPs:
                    link_dict['cp'] = CPs[i]

                nt['links'].append(link_dict)

        assert len(notes) == len(self.v2_notes_by_page_name[file_path.stem])
        assert len(notes) == self.v2_note_checksum_by_page_name[file_path.stem]

        page_dict = {'notes': list(notes.values())}
        if is_displayed_first_on_startup:
            page_dict['is_displayed_first_on_startup'] = True

        new_path = file_path.with_suffix('.json')
        json_str = json.dumps(page_dict, indent=4, ensure_ascii=False)
        new_path.write_text(json_str)
        log.info(f'Converted v2 file at {file_path} to v3 file at {new_path}')

        backup_folder.mkdir(parents=True, exist_ok=True)
        backup_file(file_path, backup_folder)
        file_path.unlink()

        return new_path

    def fix_legacy_page_internal_links(self, page: Page):

        if not isinstance(page, Page):
            raise Exception

        notes = set(self.notes(page))
        arrows = set(self.arrows(page))

        notes_updated = 0
        for note in notes:
            if note.url.is_empty():
                continue
            linked_page = self.find_one(type=Page, name=str(note.url))

            if str(note.url) == 'Imagga':
                pass
            if linked_page:
                notes_updated += 1
                note.url = linked_page.url()
                InMemoryRepository.update_one(self, note)

        if len(set(notes)) != len(notes):
            raise Exception

        if notes_updated:
            self.update_page_on_disk(page, notes, arrows)
            log.info(f'Updated {notes_updated} internal links for '
                     f'imported legacy page "{page.name}"')

        # # TODO: remove:
        # # Insert into cache
        # new_path = self.path_for_page(page)
        # old_path = new_to_old_path[new_path]
        # page_md5 = md5_by_old_name[old_path.stem]
        # _cache[page_md5] = new_path.read_text()
        # _paths_cache[page_md5] = new_path
        # if len(_cache) > MAX_CACHE:
        #     del _cache[next(iter(_cache.keys()))]
        #     del _paths_cache[next(iter(_cache.keys()))]

    def process_legacy_pages(self, previous_v_repo_entities: dict = None):
        # Collect the legacy page paths
        v2_pages = []
        v3_pages = []
        v2_bacup_folder = self.path / '__v2_legacy_pages_backup__'
        v3_backup_folder = self.path / '__v3_legacy_pages_backup__'
        for file in list(self.path.iterdir()):
            if self.is_v4_page(file) or not file.is_file():
                continue

            if file.name == '.misli_timeline_database.json':
                backup_file(file, v3_backup_folder)
                file.unlink()
                continue

            if file.suffix == '.misl':
                v2_pages.append(file)
            elif file.suffix == '.json':
                v3_pages.append(file)
            else:
                log.warning('Untracked file in the repo: %s' % file.name)
                continue

        # If there's none - return
        if not v2_pages and not v3_pages:
            return []

        legacy_pages: List[Tuple] = []  # Tuples (page, notes, arrows)

        # md5_by_old_name.clear()
        # # Try to find them in the cache
        # for page_path in copy(v2_pages):
        #     page_md5 = md5(page_path.read_bytes()).hexdigest()
        #     page_content = _cache.get(page_md5, None)
        #     md5_by_old_name[page_path.stem] = md5
        #     if page_content:
        #         v2_pages.remove(page_path)
        #         v4_path = _paths_cache[page_md5]
        #         v4_path.write_text(page_content)

        # for page_path in copy(v3_pages):
        #     page_md5 = md5(page_path.read_bytes()).hexdigest()
        #     page_content = _cache.get(page_md5, None)
        #     md5_by_old_name[page_path.stem] = md5
        #     if page_content:
        #         v3_pages.remove(page_path)
        #         v4_path = _paths_cache[page_md5]
        #         v4_path.write_text(page_content)

        # new_to_old_path.clear()  # Remove
        # v2_to_v3_path = {}

        # Process the legacy pages
        for page_path in v2_pages:
            try:
                new_path = self.convert_v2_to_v3(page_path,
                                                 backup_folder=v2_bacup_folder)
                # new_to_old_path[new_path] = page_path
                # v2_to_v3_path[page_path] = new_path
                v3_pages.append(new_path)
            except Exception as e:
                log.error(f'Exception raised when processing legacy page '
                          f'{file.path}: {e}')
                continue

        for page_path in v3_pages:
            try:
                new_path = self.convert_v3_to_v4(
                    page_path,
                    backup_folder=v3_backup_folder,
                    previous_v_repo_entities=previous_v_repo_entities)
                # if page_path in v2_to_v3_path:  # v2 to v3
                #     page_path = v2_to_v3_path[page_path]
                # new_to_old_path[new_path] = page_path

                legacy_pages.append(new_path)
            except Exception as e:
                log.error(f'Exception raised when processing legacy page '
                          f'{file.path}: {e}')
                continue

        return legacy_pages

    def checksum_imported_page_notes(self, page: Page):
        note_count_in_repo = len([
            nt for nt in self.find(parent_gid=page.gid())
            if isinstance(nt, Note)
        ])

        if page.name in self.v2_note_checksum_by_page_name:
            v2_note_count = self.v2_note_checksum_by_page_name[page.name]
            assert note_count_in_repo == v2_note_count

        if page.name in self.v3_note_checksum_by_page_name:
            v3_note_count = self.v3_note_checksum_by_page_name[page.name]
            assert note_count_in_repo == v3_note_count



from PySide6.QtCore import Qt
import pytest
import fusion
from fusion.libs import entity as entity_lib
from fusion.libs import channel as channel_lib
# from fusion.state_manager import FusionStateManager

import pamet
from pamet.desktop_app.init_config import configure_for_qt
from pamet.desktop_app.app import DesktopApp
from pamet.services.search.fuzzy import FuzzySearchService
from pamet.services.undo import UndoService
from pamet.storage.file_system.repository import FSStorageRepository
from pamet.views.window.widget import WindowWidget
from pamet.actions import window as window_actions
from pamet.actions import other as other_actions
from pamet import channels as pamet_channels


def pytest_addoption(parser):
    parser.addoption(
        '--headless',
        action='store',
        default=False,
        type=bool,
        help='Whether to show the app when running the action tests')


@pytest.fixture
def window_fixture(request, tmp_path):
    run_headless = request.config.getoption('--headless')
    # Init the app
    fusion.set_reproducible_ids(True)
    entity_lib.reset_entity_id_counter()
    channel_lib.unsibscribe_all()  # TODO: there should be a cleaner way
    fusion.fsm.reset()
    pamet.reset()

    fs_repo = FSStorageRepository.new(tmp_path, queue_save_on_change=True)
    pamet.set_sync_repo(fs_repo)

    app = DesktopApp()
    configure_for_qt(app)

    # Setup services
    pamet.set_undo_service(
        UndoService(pamet_channels.entity_change_sets_per_TLA))

    search_service = FuzzySearchService(
        pamet_channels.entity_change_sets_per_TLA)
    search_service.load_all_content()
    pamet.set_search_service(search_service)

    # Create an initial page and open the window
    start_page = other_actions.create_default_page()
    window_state = window_actions.new_browser_window()
    window = WindowWidget(initial_state=window_state)
    if run_headless:
        window.setAttribute(Qt.WA_DontShowOnScreen)

    window_actions.new_browser_tab(window_state, start_page)
    window.showMaximized()
    fusion.main_loop().process_events()

    yield window
    app.shutdown()



from typing import List
from fusion.libs.entity.change import Change
from fusion.libs.entity import Entity
from fusion.libs.channel import Channel
from fusion.storage.in_memory_repository import InMemoryRepository
from fusion.storage.repository import Repository
from pamet.storage.base_repository import PametRepository


class PersistenceManager(Repository):
    """TODO: finish the implementation.
    Receives changesets on the input_channel and sends them to the async repo.
    Then receives the successfull and unsuccessfull saves on the respective
    channels and applies some policy to handle the unsuccessful ones (initially
    just throw an error or something)

    Also the state of the class can be connected to the gui to reflect the
    save status. donno how yet"""

    def __init__(self,
                 in_channel: Channel,
                 sync_repo: PametRepository,
                 async_repo: Repository = None):
        self.in_channel = in_channel
        self.sync_repo = sync_repo
        self.async_repo = async_repo
        self._change_sets = []

        # Create out/save channels
        # self._out_channel = Channel('CHANGES_FOR_SAVING')
        # self._saved_changes_channel = Channel('SAVED_CHANGES')
        self.changes_for_async_save_channel = Channel(
            '__CHANGES_FOR_ASYNC_SAVE__')
        self.unsuccessfully_saved_changes_channel = Channel(
            '__UNSUCCESSFUL_SAVES__')
        self.successfully_saved_changes_channel = Channel(
            '__SUCCESSFUL_SAVES__')

        # Connect the channels
        self.in_channel.subscribe(self.queue_save_for_changeset)
        self.repository.set_save_channel(self._saved_changes_channel)
        self.successfully_saved_changes_channel.subscribe(
            self.handle_saved_changes)

        # crud methods with buffering and calling the repo when the object is
        # not buffered

    def handle_saved_changes(self, changes: List[Change]):
        pass

    def queue_save_for_changeset(self, changes: List[Change]):
        self._change_sets.append(changes)



from typing import Generator, Union
from fusion.libs.entity.change import Change
from fusion.util import current_time
from fusion.libs.channel import Channel
from fusion.storage.repository import Repository

from pamet.model.arrow import Arrow
from pamet.model.note import Note
from pamet.model.page import Page
from pamet.model.page_child import PageChild


class PametRepository(Repository):

    def __init__(self) -> None:
        self.raw_entity_changes_channel = None
        self.default_page_id = None

    def set_change_channel(self, change_channel: Channel):
        self.raw_entity_changes_channel = change_channel

    # -------------Pages CRUD-------------
    def insert_page(self, page_: Page) -> Change:
        change = self.insert_one(page_)
        if self.raw_entity_changes_channel:
            self.raw_entity_changes_channel.push(change)
        return change

    def remove_page(self, page_: Page) -> Change:
        change = self.remove_one(page_)
        if self.raw_entity_changes_channel:
            self.raw_entity_changes_channel.push(change)
        return change

    def update_page(self, page_: Page) -> Change:
        old_page = self.find_one(id=page_.id)
        if not old_page:
            raise Exception('Can not update missing page.')

        if page_.name != old_page.name:
            page_.datetime_modified = current_time()

        change = self.update_one(page_)
        if self.raw_entity_changes_channel:
            self.raw_entity_changes_channel.push(change)
        return change

    def pages(self, **filter) -> Generator[Page, None, None]:
        filter['type'] = Page
        return self.find(**filter)

    def page(self, page_gid: str | tuple) -> Union[Page, None]:
        if not page_gid:
            return None
        return self.find_one(gid=page_gid)

    def insert_child(self, child: PageChild, page: Page = None) -> Change:
        if page and page.id != child.page_id:
            child = child.with_id(page.id, child.own_id)
        change = self.insert_one(child)
        if self.raw_entity_changes_channel:
            self.raw_entity_changes_channel.push(change)
        return change

    # -------------Notes CRUD-------------
    def insert_note(self, note: Note, page: Page = None) -> Change:
        return self.insert_child(note, page)

    def update_note(self, note_: Note) -> Change:
        old_note = self.find_one(gid=note_.gid())
        if not old_note:
            raise Exception('Can not update missing note.')

        if note_.content != old_note.content:
            note_.datetime_modified = current_time()

        change = self.update_one(note_)
        if self.raw_entity_changes_channel:
            self.raw_entity_changes_channel.push(change)
        return change

    def remove_note(self, note_: Note) -> Change:
        change = self.remove_one(note_)
        if self.raw_entity_changes_channel:
            self.raw_entity_changes_channel.push(change)
        return change

    def notes(self, page_: Page | str) -> Generator[Note, None, None]:
        page_gid = page_.gid() if isinstance(page_, Page) else page_
        return self.find(parent_gid=page_gid, type=Note)

    def note(self, page_: Page | str, note_own_id: str):
        page_gid = page_.gid() if isinstance(page_, Page) else page_
        return self.find_one(gid=(page_gid, note_own_id))

    # def notes(self, **kwargs) -> Generator[Note, None, None]:
    #     if 'type' in kwargs:
    #         raise Exception
    #     return self.find(type=Note, **kwargs)

    # def note(self, **kwargs):
    #     if 'type' in kwargs:
    #         raise Exception
    #     return self.find_one(type=Note, **kwargs)

    # -------------Arrow CRUD-------------
    def insert_arrow(self, arrow_: Arrow, page: Page = None) -> Change:
        return self.insert_child(arrow_, page)

    def update_arrow(self, arrow_: Arrow) -> Change:
        old_arrow = self.find_one(gid=arrow_.gid())
        if not old_arrow:
            raise Exception('Can not update missing arrow')

        change = self.update_one(arrow_)
        if self.raw_entity_changes_channel:
            self.raw_entity_changes_channel.push(change)
        return change

    def remove_arrow(self, arrow_: Arrow) -> Change:
        change = self.remove_one(arrow_)
        if self.raw_entity_changes_channel:
            self.raw_entity_changes_channel.push(change)
        return change

    def arrows(self, page_: Page | str):
        page_gid = page_.gid() if isinstance(page_, Page) else page_
        return self.find(parent_gid=page_gid, type=Arrow)

    def arrow(self, page_: Page | str, arrow_id: str):
        page_gid = page_.gid() if isinstance(page_, Page) else page_
        return self.find_one(gid=(page_gid, arrow_id))

    # Other
    def apply_change(self, change: Change):
        last_state = change.last_state()

        if change.is_create():
            self.insert_one(last_state)
        elif change.is_update():
            self.update_one(last_state)
        elif change.is_delete():
            self.remove_one(last_state)

        if self.raw_entity_changes_channel:
            self.raw_entity_changes_channel.push(change)

    def write_to_disk(self):
        # TODO: This should be redundant when I implement the
        # Persistence Manager
        pass

    def default_page(self):
        return self.page(self.default_page_id)

    def set_default_page(self, new_page: Page):
        self.default_page_id = new_page.id



from pathlib import Path
from fastapi import FastAPI, HTTPException, Request
from fastapi.middleware.cors import CORSMiddleware
from fusion.libs.entity.change import Change

from pamet.storage.file_system.repository import FSStorageRepository

REPO_DIR = Path.home() / 'pamet_server_repo'
DEBUG = True

app = FastAPI()

origins = [
    'http://localhost',
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=['*'],
    allow_headers=['*'],
)

repo = FSStorageRepository(REPO_DIR)
repo.load_all_pages()


@app.get('/api/page/{page_id}/children')
def get_children(page_id: str):
    page = repo.page(page_id)
    if not page:
        raise HTTPException(status_code=404, detail='Page not found')

    note_dicts = []
    for note in repo.notes(page_id):
        note_dict = note.asdict()
        note_dict['type'] = type(note).__name__
        note_dicts.append(note_dict)

    arrow_dicts = []
    for arrow in repo.arrows(page_id):
        arrow_dict = arrow.asdict()
        arrow_dict['type'] = type(arrow).__name__
        arrow_dicts.append(arrow_dict)

    return {
        'notes': note_dicts,
        'arrows': arrow_dicts,
    }


@app.post('/api/page/{page_id}/changes')
async def apply_changes(request: Request):
    json_payload = await request.json()
    change_dicts = json_payload.get('changes')
    if not change_dicts:
        raise HTTPException(status_code=400, detail='No changes provided')

    try:
        changes = [Change.from_dict(change) for change in change_dicts]
    except Exception as e:
        err_message = f'Invalid change'
        if DEBUG:
            err_message += f': {e}'
        raise HTTPException(status_code=400, detail=err_message)

    failed_changes = []
    successfull_changes = []
    for change_dict in change_dicts:
        try:
            change = Change.from_dict(change_dict)
        except Exception as e:
            err_message = 'Invalid change'
            if DEBUG:
                err_message += f': {e}'
            failed_changes.append({
                'change': change_dict,
                'error': err_message,
            })
            continue

        try:
            repo.apply_change(change)
        except Exception as e:
            err_message = 'Failed to apply change'
            if DEBUG:
                err_message += f': {e}'
            failed_changes.append({
                'change': change_dict,
                'error': err_message,
            })
            continue

        successfull_changes.append(change.id)

    if not failed_changes:
        return HTTPException(status_code=207, detail={
            'applied_changes': successfull_changes,
            'errors': failed_changes,
        })
    else:
        return {
            'applied_changes': successfull_changes,
        }




from pamet.views.search_bar.widget import SearchBarWidget
from fusion.libs.state import ViewState, view_state_type
import pamet


@view_state_type
class SemanticSearchBarWidgetState(ViewState):
    pass


class SemanticSearchBarWidget(SearchBarWidget):

    @property
    def search_service(self):
        return pamet.semantic_search_service()



from copy import copy
from typing import List, Tuple, Union

import math
from PySide6.QtCore import QObject, QPointF, QRectF
from PySide6.QtGui import QColor, QPainter, QPainterPath

from fusion import fsm
from fusion.util.point2d import Point2D
from fusion.util.rectangle import Rectangle
from fusion.libs.entity.change import Change
from fusion.platform.qt_widgets import bind_and_apply_state
from fusion.view import View
from fusion.libs.state import ViewState, view_state_type
from fusion.logging import get_logger
import pamet
from pamet.desktop_app import selection_overlay_qcolor
from pamet.constants import ARROW_EDGE_RAIDUS, ARROW_SELECTION_THICKNESS_DELTA, CONTROL_POINT_RADIUS, POTENTIAL_EDGE_RADIUS

from pamet.model.arrow import BEZIER_CUBIC, Arrow, ArrowAnchorType

log = get_logger(__name__)

CONTROL_POINT_DEBUG_VISUALS = False

TAIL = 'tail'
HEAD = 'head'
ARROW_HAND_LENGTH = 20
ARROW_HAND_ANGLE = math.radians(25)
CP_BASE_DISTANCE = 80
CP_DIST_SEGMENT_ADJUST_K = 0.1


def special_sigmoid(x: float) -> float:
    return 1 / (1 + math.exp(-x / (CP_BASE_DISTANCE / 2) + 5))


@view_state_type
class ArrowViewState(Arrow, ViewState):

    @property
    def arrow_gid(self):  # TODO: remove
        return self.id

    def get_arrow(self):
        return pamet.find_one(gid=self.arrow_gid)

    def update_from_arrow(self, arrow: Arrow):
        arrow_dict = arrow.asdict()
        assert self.id == arrow_dict.pop('id')
        self.replace(**arrow_dict)


class ArrowView(View):

    def intersects_rect(self, rect: Rectangle) -> bool:
        raise NotImplementedError

    def intersects_circle(self, center: Point2D, radius: float) -> bool:
        raise NotImplementedError


class ArrowWidget(QObject, ArrowView):

    def __init__(self, initial_state: ArrowViewState = None, parent=None):
        QObject.__init__(self, parent)
        ArrowView.__init__(self, initial_state)
        self.map_page_view = parent
        self._anchor_subs_by_note_id = {}
        self._cached_curves = None
        self._cached_path = None

        bind_and_apply_state(self,
                             initial_state,
                             on_state_change=self.on_state_change)
        self.destroyed.connect(lambda: self.unsubscribe_all())

    def unsubscribe_all(self):
        for note_id in copy(self._anchor_subs_by_note_id):
            self.unsubscribe_from_anchor(note_id)

    def update_anchor_subscriptions(self):
        state = self.state()
        new_ids = set()
        if state.tail_note_id:
            new_ids.add(state.tail_note_id)
        if state.head_note_id:
            new_ids.add(state.head_note_id)

        old_ids = set(self._anchor_subs_by_note_id.keys())
        new_subs = new_ids - old_ids
        subs_for_removal = old_ids - new_ids

        for note_id in subs_for_removal:
            self.unsubscribe_from_anchor(note_id)

        for note_id in new_subs:
            self.subscribe_to_anchor(note_id)

    def subscribe_to_anchor(self, note_own_id: str):
        # Clear the previous subscription
        if note_own_id in self._anchor_subs_by_note_id:
            self.unsubscribe_from_anchor(note_own_id)

        map_page_state = self.map_page_view.state()
        note_view_state = map_page_state.view_state_for_note_own_id(note_own_id)
        if not note_view_state:
            raise Exception('Anchor note with own id {note_own_id} not found')

        sub = fsm.state_changes_per_TLA_by_view_id.subscribe(
            handler=self.handle_anchor_note_view_state_change,
            index_val=note_view_state.view_id)
        self._anchor_subs_by_note_id[note_own_id] = sub

    def unsubscribe_from_anchor(self, note_own_id: str):
        if note_own_id not in self._anchor_subs_by_note_id:
            return
        sub = self._anchor_subs_by_note_id.pop(note_own_id)
        sub.unsubscribe()

    def handle_anchor_note_view_state_change(self, change):
        if change.is_delete():
            return  # The arrow should be deleted along the deleted note
        self.update_cached_path()

    def on_state_change(self, change: Change):
        # Update anchor note subscriptions
        if change.updated.tail_note_id or change.updated.head_note_id:
            self.update_anchor_subscriptions()

        if not change.is_delete():
            self.update_cached_path()

    def calculate_terminal_cp(self, terminal_point: Point2D,
                              adjacent_point: Point2D,
                              control_point_distance: float,
                              anchor_type: ArrowAnchorType):
        if anchor_type == ArrowAnchorType.NONE:
            k = control_point_distance / terminal_point.distance_to(
                adjacent_point)
            return terminal_point + (adjacent_point - terminal_point) * k
        elif anchor_type == ArrowAnchorType.MID_LEFT:
            return terminal_point - Point2D(control_point_distance, 0)
        elif anchor_type == ArrowAnchorType.TOP_MID:
            return terminal_point - Point2D(0, control_point_distance)
        elif anchor_type == ArrowAnchorType.MID_RIGHT:
            return terminal_point + Point2D(control_point_distance, 0)
        elif anchor_type == ArrowAnchorType.BOTTOM_MID:
            return terminal_point + Point2D(0, control_point_distance)
        else:
            raise Exception

    def infer_arrow_anchor_type(self, adjacent_point: Point2D,
                                note_rect: Rectangle):
        # If the adjacent point is to the left or right - set a side anchor
        if adjacent_point.x() < note_rect.left():
            return ArrowAnchorType.MID_LEFT
        elif adjacent_point.x() > note_rect.right():
            return ArrowAnchorType.MID_RIGHT
        else:
            # If the point is directly above or below the note - set a
            # top/bottom anchor
            if adjacent_point.y() < note_rect.top():
                return ArrowAnchorType.TOP_MID
            elif adjacent_point.y() > note_rect.bottom():
                return ArrowAnchorType.BOTTOM_MID
            else:
                # If the adjacent point is inside the note_rect - set either
                # a top or bottom anchor (arbitrary, its ugly either way)
                if adjacent_point.y() < note_rect.center():
                    return ArrowAnchorType.TOP_MID
                else:
                    return ArrowAnchorType.BOTTOM_MID

    def cp_distance_for_segment(self, first_point, second_point):
        dist = first_point.distance_to(second_point)

        return (special_sigmoid(dist) * CP_BASE_DISTANCE +
                CP_DIST_SEGMENT_ADJUST_K * dist)

    def bezier_cubic_curves_params(
        self,
        tail_point: Point2D = None,
        head_point: Point2D = None
    ) -> List[Tuple[Point2D, Point2D, Point2D, Point2D]]:
        """ Returns a list of curve parameters. Each set of parameters is a
        tuple in the form:
        (start_point, first_control_point, second_control point, end_point)
        The parameters tail_point and head_point can override the corresponding
        internal variables in order to accomodate for anchor position changes.
        """
        tail_point = tail_point or self.tail_point
        head_point = head_point or self.head_point
        curves = []
        state = self.state()

        # Handle the first control point of the first curve
        if state.mid_points:
            second_point = state.mid_points[0]
        else:
            second_point = head_point

        # If the anchor type is AUTO - infer it
        if (state.has_tail_anchor()
                and state.tail_anchor_type == ArrowAnchorType.AUTO):
            tail_note = pamet.note(state.page_id, state.tail_note_id)
            note_view = self.map_page_view.note_widget_by_note_gid(
                tail_note.gid())

            tail_anchor_type = self.infer_arrow_anchor_type(
                second_point, note_view.rect())
        else:
            tail_anchor_type = state.tail_anchor_type

        control_point_distance = self.cp_distance_for_segment(
            tail_point, second_point)

        first_cp = self.calculate_terminal_cp(tail_point, second_point,
                                              control_point_distance,
                                              tail_anchor_type)

        # Add the second control point for the first curve
        # and all the middle curves (if any), excluding the last control point
        prev_point: Point2D = tail_point
        for idx, current_point in enumerate(state.mid_points):
            # if idx == 0:
            #     prev_point: Point2D = tail_point

            # If we're at the last point
            if (idx + 1) == len(state.mid_points):
                next_point = head_point
                # should_finish = True
            else:
                next_point = state.mid_points[idx + 1]

            # Calculate alpha (see the schematic for details)
            a = current_point.distance_to(next_point)
            b = prev_point.distance_to(current_point)
            # c = prev_point.distance_to(next_point)
            # beta = math.acos((a**2 + b**2 - c**2) / (a * b * 2))
            dA = prev_point - current_point
            dB = next_point - current_point
            gamma = math.atan2(dA.y(), dA.x())
            theta = math.atan2(dB.y(), dB.x())
            if gamma < 0:
                gamma += math.pi * 2
            if theta < 0:
                theta += math.pi * 2
            beta = (math.pi * 2 + theta - gamma) if gamma > theta else (theta -
                                                                        gamma)
            alpha = math.pi / 2 - beta / 2  # In radians 90 - beta/2

            control_point_distance = self.cp_distance_for_segment(
                prev_point, current_point)

            # Calculate the second control point for the first curve
            if b == 0:
                b = 0.0001
            k = control_point_distance / b
            z_prim: Point2D = current_point + k * (prev_point - current_point)
            second_control_point = z_prim.rotated(alpha, current_point)
            # second_control_point = z_prim
            # second_control_point = current_point + Point2D(50, 50)

            # Save the params
            curves.append(
                (prev_point, first_cp, second_control_point, current_point))

            # Calculate the first control point for the second curve
            # (mirrors the second cp of the last curve)
            control_point_distance = self.cp_distance_for_segment(
                current_point, next_point)
            if a == 0:
                a = 0.0001
            k = control_point_distance / a
            q_prim = current_point + k * (next_point - current_point)
            first_cp = q_prim.rotated(-alpha, current_point)

            prev_point = current_point

        # If the anchor type is AUTO - infer it
        if (state.has_head_anchor()
                and state.head_anchor_type == ArrowAnchorType.AUTO):
            head_note = pamet.note(state.page_id, state.head_note_id)
            note_view = self.map_page_view.note_widget_by_note_gid(
                head_note.gid())

            head_anchor_type = self.infer_arrow_anchor_type(
                second_point, note_view.rect())
        else:
            head_anchor_type = state.head_anchor_type

        if not state.mid_points:
            # Add the second control point for the last curve
            last_control_point = self.calculate_terminal_cp(
                head_point, tail_point, control_point_distance,
                head_anchor_type)
        else:
            # Add the second control point for the last curve
            last_control_point = self.calculate_terminal_cp(
                head_point, current_point, control_point_distance,
                head_anchor_type)

        curves.append((prev_point, first_cp, last_control_point, head_point))

        return curves

    def update_cached_path(self):
        state: ArrowViewState = self.state()
        if state.has_tail_anchor():
            tail_note = pamet.note(state.page_id, state.tail_note_id)
            nv_state = self.map_page_view.get_note_view_state_for_note_gid(
                tail_note.gid())
            if not nv_state:
                raise Exception
            if state.tail_anchor:
                tail_anchor_pos = nv_state.arrow_anchor(state.tail_anchor_type)
            else:
                raise NotImplementedError
        elif state.tail_point:
            tail_anchor_pos = state.tail_point
        else:
            # there should be either a position or an anchor set
            # if on init there's acceptable cases without - just return here
            return
            raise Exception

        if state.head_note_id:
            head_note = pamet.note(state.page_id, state.head_note_id)
            nv_state = self.map_page_view.get_note_view_state_for_note_gid(
                head_note.gid())
            if not nv_state:
                raise Exception
            if state.head_anchor:
                head_anchor_pos = nv_state.arrow_anchor(state.head_anchor_type)
            else:
                raise NotImplementedError
        elif state.head_point:
            head_anchor_pos = state.head_point
        else:
            # there should be either a position or an anchor set
            # if on init there's acceptable cases without - just return here
            return

        if tail_anchor_pos == head_anchor_pos:
            self._cached_path = None
            return

        if state.line_function_name == BEZIER_CUBIC:
            curves = self.bezier_cubic_curves_params(tail_anchor_pos,
                                                     head_anchor_pos)
            if not curves:
                raise Exception

            self._cached_curves = curves

            start_point = curves[0][0]
            start_point = QPointF(*start_point.as_tuple())
            self._cached_path = QPainterPath(start_point)
            # curves.extend(reversed([reversed(c) for c in curves]))
            for first_point, first_cp, second_cp, second_point in curves:
                first_point = QPointF(*first_point.as_tuple())
                first_cp = QPointF(*first_cp.as_tuple())
                second_cp = QPointF(*second_cp.as_tuple())
                second_point = QPointF(*second_point.as_tuple())

                self._cached_path.cubicTo(first_cp, second_cp, second_point)

            # Add the same path reversed in order to have proper filling and
            # intersections (otherwise it draws a start-end line to close the
            # polygon)
            self._cached_path.addPath(self._cached_path.toReversed())
        else:
            raise NotImplementedError

    def draw_arrow_path_and_heads(self, painter: QPainter):
        painter.drawPath(self._cached_path)

        # Draw the arrow head
        point_at_end = self._cached_curves[-1][-2]
        # point_at_end = self._cached_path.pointAtPercent(0.97)
        point_at_end = Point2D(point_at_end.x(), point_at_end.y())
        end_point = self._cached_curves[-1][-1]
        # end_point = self._cached_path.pointAtPercent(1)  # 100%
        end_point = Point2D(end_point.x(), end_point.y())

        normalized_vec = (end_point -
                          point_at_end) / point_at_end.distance_to(end_point)
        arrow_hand_base = end_point - ARROW_HAND_LENGTH * normalized_vec
        hand_end = arrow_hand_base.rotated(ARROW_HAND_ANGLE, end_point)
        hand_end2 = arrow_hand_base.rotated(-ARROW_HAND_ANGLE, end_point)

        end_point = QPointF(*end_point.as_tuple())
        hand_end = QPointF(*hand_end.as_tuple())
        hand_end2 = QPointF(*hand_end2.as_tuple())
        painter.drawLine(end_point, hand_end)
        painter.drawLine(end_point, hand_end2)

    def render(self, painter: QPainter, draw_selection_overlay: bool,
               draw_control_points: bool):
        if not self._cached_path or not self._cached_path.elementCount():
            log.warning('Render called, but _cached_path is empty')
            return

        state: ArrowViewState = self.state()

        # A hacky curve update in case that a render has been called before
        # the update_cache_state has been invoked from the state update
        # it should be a problem specific to the custom drawing setup
        if len(self._cached_curves) != (len(state.edge_indices()) - 1):
            self.update_cached_path()

        pen = painter.pen()
        pen.setColor(QColor(*state.get_color().to_uint8_rgba_list()))
        pen.setWidthF(state.line_thickness)
        painter.setPen(pen)

        self.draw_arrow_path_and_heads(painter)

        if draw_selection_overlay:
            pen.setColor(selection_overlay_qcolor)
            pen.setWidthF(state.line_thickness +
                          ARROW_SELECTION_THICKNESS_DELTA)
            painter.setPen(pen)
            self.draw_arrow_path_and_heads(painter)

        if draw_control_points:
            # The tail is at idx 0, the head at the last, the midpoints
            # also have integer indices and the potential new midpoints have .5
            # indices

            for idx in state.edge_indices():
                edge_point = QPointF(*self.edge_point_pos(idx).as_tuple())
                painter.drawEllipse(edge_point, CONTROL_POINT_RADIUS,
                                    CONTROL_POINT_RADIUS)

            for idx in state.potential_edge_indices():
                edge_point = QPointF(*self.edge_point_pos(idx).as_tuple())
                painter.drawEllipse(edge_point, POTENTIAL_EDGE_RADIUS,
                                    POTENTIAL_EDGE_RADIUS)

        if CONTROL_POINT_DEBUG_VISUALS:
            for curve_idx, curve in enumerate(self._cached_curves):

                c1, c2 = (QPointF(*p.as_tuple()) for p in curve[1:3])
                painter.drawEllipse(c1, 10, 10)
                painter.drawEllipse(c2, 10, 10)

                if (curve_idx + 1) < len(self._cached_curves):
                    next_curve = self._cached_curves[curve_idx + 1]
                    nc1, nc2 = (QPointF(*p.as_tuple())
                                for p in next_curve[1:3])
                    painter.drawLine(c2, nc1)

    def edge_point_pos(self, edge_index: float) -> Point2D:
        """Returns the edge point position for the given index. Those
        include the tail, midpoints and head.

        The indeces are not integers, since the suggested new edge points
        are denoted with non-whole indices (e.g. 0.5, 1.5 etc.).
        More over the tail index is 0, and the head index is the last one.
        """

        if edge_index == 0:
            return self._cached_curves[0][0]
        elif edge_index % 1 == 0:
            curve_idx = int(edge_index - 1)
            curve = self._cached_curves[curve_idx]
            _, _, _, point = curve
            return copy(point)
        else:  # A non-whole index - a potential cp is requested
            curve_idx = int(edge_index)
            curve = self._cached_curves[curve_idx]
            p1, cp1, cp2, p2 = (QPointF(*p.as_tuple()) for p in curve)

            path = QPainterPath(p1)
            path.cubicTo(cp1, cp2, p2)
            potential_cp = path.pointAtPercent(0.5)
            return Point2D(potential_cp.x(), potential_cp.y())

    def edge_at(self, position: Point2D) -> Union[float, None]:
        real_pos = self.map_page_view.state().unproject_point(position)
        indices = self.state().all_edge_indices()
        for idx in indices:
            point = self.edge_point_pos(idx)

            if idx % 1 == 0:  # If it's a real edge
                radius = ARROW_EDGE_RAIDUS
            else:
                radius = POTENTIAL_EDGE_RADIUS
            if point.distance_to(real_pos) <= radius:
                return idx
        return None

    def intersects_circle(self, center: Point2D, radius: float) -> bool:
        if not self._cached_path:
            return False

        path = QPainterPath()
        path.addEllipse(QPointF(*center.as_tuple()), radius, radius)
        return self._cached_path.intersects(path)

    def intersects_rect(self, rect: Rectangle) -> bool:
        if not self._cached_path:
            return False

        selector_rect = QRectF(*rect.as_tuple())
        return self._cached_path.intersects(selector_rect)



from dataclasses import field

from fusion.libs.state import view_state_type, ViewState
from fusion.view import View


@view_state_type
class WindowViewState(ViewState):
    title: str = ''
    current_tab_id: str = ''
    tab_states: list = field(default_factory=list)
    command_view_state: View = None

    def __repr__(self) -> str:
        return (f'<WindowViewState title={self.title} {self.current_tab_id=}'
                f' {len(self.tab_states)=}>')



from fusion.libs.state import view_state_type
from pamet.model.script_note import ScriptNote
from pamet.note_view_lib import register_note_view_type
from pamet.views.note.base_edit.view_state import NoteEditViewState
from pamet.views.note.base_edit.widget import BaseNoteEditWidget
from pamet.views.note.script.props_widget import ScriptNotePropsWidget


@view_state_type
class ScriptNoteEditViewState(NoteEditViewState, ScriptNote):
    pass


@register_note_view_type(state_type=ScriptNoteEditViewState,
                         note_type=ScriptNote,
                         edit=True)
class ScriptNoteEditWidget(BaseNoteEditWidget):

    def __init__(self, parent, initial_state: ScriptNoteEditViewState):
        BaseNoteEditWidget.__init__(self, parent, initial_state)

        self.script_props_widget = ScriptNotePropsWidget(self)
        self.ui.centralAreaWidget.layout().addWidget(
            self.script_props_widget)



from PySide6.QtWidgets import QWidget
from pamet import commands

from pamet.actions import tab as tab_actions
from pamet.views.context_menu.widget import ContextMenuWidget

from .ui_widget import Ui_CornerWidget


class CornerWidget(QWidget):

    def __init__(self, parent):
        super().__init__(parent)

        self.window = parent
        self.ui = Ui_CornerWidget()
        self.ui.setupUi(self)

        self.ui.navigationBackButton.clicked.connect(
            lambda: tab_actions.navigation_back(self.window.current_tab().
                                                state()))

        self.ui.navigationForwardButton.clicked.connect(
            lambda: tab_actions.navigation_forward(self.window.current_tab().
                                                   state()))

        self.ui.navigationToggleButton.clicked.connect(
            lambda: tab_actions.navigation_toggle_last(self.window.current_tab(
            ).state()))

        self.ui.menuButton.clicked.connect(self.open_main_menu)

    def open_main_menu(self):
        entries = {
            'Page properties': commands.open_page_properties,
            'Open user settings (JSON)': commands.open_user_settings_json,
            'Open repo settings (JSON)': commands.open_repo_settings_json,
            'Export as HTML page /*buggy*': commands.export_as_web_page,
        }
        context_menu = ContextMenuWidget(self, entries=entries)
        context_menu.popup(self.ui.menuButton.rect().bottomRight())



from PySide6.QtGui import QIcon
from pamet.util import resource_path


class PametQtWidgetsCachedIcons:
    trash: QIcon = None
    link: QIcon = None
    plus: QIcon = None
    text: QIcon = None
    image: QIcon = None

    def load_all(self):
        self.trash = QIcon(str(resource_path('icons/delete-bin-line.svg')))
        self.link = QIcon(str(resource_path('icons/link.svg')))
        self.plus = QIcon(str(resource_path('icons/plus.svg')))
        self.text = QIcon(str(resource_path('icons/text.svg')))
        self.text.addFile(str(resource_path('icons/text.svg')),
                          mode=QIcon.Disabled)
        self.image = QIcon(str(resource_path('icons/image-line.svg')))
        self.more = QIcon(str(resource_path('icons/more-2-line.svg')))



from fusion.libs.channel import Channel

# Entity changeset per top-level action (TLA), meaning the channel transmits
# change-set objects (lists or sets) containing the changes made in a TLA
# Note: A TLA is a non-nested action.
entity_change_sets_per_TLA = Channel('__ENTITY_CHANGESETS_PER_TLA__')

entity_changes_by_id = Channel('__ENTITY_CHANGES_BY_ID__',
                               index_key=lambda x: x.last_state().id)
entity_changes_by_parent_gid = Channel(
    '__ENTITY_CHANGE_BY_PARENT_GID__',
    index_key=lambda x: x.last_state().parent_gid())
# page_changes = Channel('__PAGE_CHANGES__',
#                        filter_key=lambda c: isinstance(c.last_state(), Page))
